# lsqfitgp/_gvarext.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools
import textwrap
import string
import contextlib

import jax
import gvar
from jax import numpy as jnp
import numpy

from . import _signature

__all__ = [
    'jacobian',
    'from_jacobian',
    'gvar_gufunc',
    'switchgvar',
]

def _getsvec(x):
    """
    Get the sparse vector of derivatives of a GVar.
    """
    if isinstance(x, gvar.GVar):
        return x.internaldata[1]
    else:
        return gvar.svec(0)

def _merge_svec(gvlist, start=None, stop=None):
    if start is None:
        return _merge_svec(gvlist, 0, len(gvlist))
    n = stop - start
    if n <= 0:
        return gvar.svec(0)
    if n == 1:
        return _getsvec(gvlist[start])
    left = _merge_svec(gvlist, start, start + n // 2)
    right = _merge_svec(gvlist, start + n // 2, stop)
    return left.add(right, 1, 1)

def jacobian(g):
    """
    Extract the jacobian of gvars w.r.t. primary gvars.
    
    Parameters
    ----------
    g : array_like
        An array of numbers or gvars.
    
    Returns
    -------
    jac : array
        The shape is g.shape + (m,), where m is the total number of primary
        gvars that g depends on.
    indices : (m,) int array
        The indices that map the last axis of jac to primary gvars in the
        global covariance matrix.

    See also
    --------
    from_jacobian
    """
    g = numpy.asarray(g)
    v = _merge_svec(g.flat)
    indices = v.indices()
    jac = numpy.zeros((g.size, len(indices)), float)
    for i, x in enumerate(g.flat):
        v = _getsvec(x)
        ind = numpy.searchsorted(indices, v.indices())
        jac[i, ind] = v.values()
    jac = jac.reshape(g.shape + indices.shape)
    return jac, indices

def from_jacobian(mean, jac, indices):
    """
    Create new gvars from a jacobian w.r.t. primary gvars.
    
    Parameters
    ----------
    mean : array_like
        An array of numbers with the means of the new gvars.
    jac : mean.shape + (m,) array
        The derivatives of each new gvar w.r.t. m primary gvars.
    indices : (m,) int array
        The indices of the primary gvars.
    
    Returns
    -------
    g : mean.shape array
        The new gvars.

    See also
    --------
    jacobian
    """
    cov = gvar.gvar.cov
    mean = numpy.asarray(mean)
    shape = mean.shape
    mean = mean.flat
    jac = numpy.asarray(jac)
    jac = jac.reshape(len(mean), len(indices))
    g = numpy.zeros(len(mean), object)
    for i, jacrow in enumerate(jac):
        der = gvar.svec(len(indices))
        der._assign(jacrow, indices)
        g[i] = gvar.GVar(mean[i], der, cov)
    return g.reshape(shape)

def tabulate_together(*gs, headers=True, offset='', ndecimal=None, keys=None):
    """
    
    Format a table comparing side by side various collections of gvars.
    
    Parameters
    ----------
    *gs : sequence of arrays or dictionaries of gvars
        The variables to be tabulated. The structures of arrays and dictionaries
        must match.
    headers : bool or sequence of strings
        If True (default), add automatically an header. If False, don't add an
        header. If a sequence with length len(gs) + 1, it contains the column
        names for the keys/indices and for each set of variables.
    offset : str
        Prefix to each line, default empty.
    ndecimal : int, optional
        Number of decimal places. If not specified (default), keep two error
        digits.
    keys : sequence, optional
        If ``gs`` are dictionaries, a subset of keys to be extracted from each
        dictionary. Ignored if they are arrays.
    
    Examples
    --------
    >>> print(tabulate_together(gvar.gvar(dict(a=1)), gvar.gvar(dict(a=2))))
    key/index   value1   value2
    ---------------------------
            a    1 (0)    2 (0)
    
    See also
    --------
    gvar.tabulate
    
    """
    if not gs:
        return ''
    gs = [g if hasattr(g, 'keys') else numpy.asarray(g) for g in gs]
    assert all(hasattr(g, 'keys') for g in gs) or all(not hasattr(g, 'keys') for g in gs)
    if keys is not None and hasattr(gs[0], 'keys'):
        gs = [{k: g[k] for k in keys} for g in gs]
    g0 = gs[0]
    if hasattr(g0, 'keys'):
        assert all(set(g.keys()) == set(g0.keys()) for g in gs[1:])
        gs = [{k: g[k] for k in g0} for g in gs]
    else:
        assert all(g.shape == g0.shape for g in gs[1:])
        if g0.shape == ():
            gs = [{'--': g} for g in gs]
    tables = [
        _splittable(gvar.tabulate(g, headers=['@', ''], ndecimal=ndecimal))
        for g in gs
    ]
    columns = list(tables[0]) + [t[1] for t in tables[1:]]
    if not hasattr(headers, '__len__'):
        if headers:
            headers = ['key/index'] + [f'value{i+1}' for i in range(len(gs))]
        else:
            headers = None
    else:
        assert len(headers) == len(columns)
    if headers is not None:
        columns = (_head(col, head) for col, head in zip(columns, headers))
    return textwrap.indent(_join(columns), offset)

def _splittable(table):
    lines = table.split('\n')
    header = lines[0]
    col = header.find('@') + 1
    contentlines = lines[2:]
    col1 = '\n'.join(line[:col] for line in contentlines)
    col2 = '\n'.join(line[col:] for line in contentlines)
    return col1, col2

def _head(col, head):
    head = str(head)
    width = col.find('\n')
    if width < 0:
        width = len(col)
    hwidth = len(head)
    if hwidth > width:
        col = textwrap.indent(col, (hwidth - width) * ' ')
    else:
        head = (width - hwidth) * ' ' + head
    return head + '\n' + len(head) * '-' + '\n' + col

def _join(cols):
    split = (col.split('\n') for col in cols)
    return '\n'.join(''.join(lines) for lines in zip(*split))

def gvar_gufunc(func, *, signature=None):
    """

    Wraps a jax-traceable generalized ufunc with one argument to support gvars.

    Parameters
    ----------
    func : callable
        A function from one array to one array. It must be a generalized ufunc,
        and differentiable one time with `jax`.
    signature : str, optional
        The signature of the generalized ufunc. If not specified, it is assumed
        to be scalar to scalar (normal ufunc).

    Returns
    -------
    decorated_func : callable
        A function that, in addition to numerical arrays, accepts gvars and
        returns gvars.

    See also
    --------
    numpy.vectorize

    """

    # parse signature
    if signature is None:
        signature = '()->()'
    sig = _signature.Signature(signature)
    inp, = sig.incores
    out, = sig.outcores
    jac_sig = _signature.Signature.from_tuples([inp], [out + inp])
    
    # make jacobian function
    deriv = jnp.vectorize(jax.jacfwd(func), signature=jac_sig.signature)

    # get indices for summation
    ninp = len(inp)
    nout = len(out)
    head_indices = '...'
    out_indices = string.ascii_letters[:nout]
    in_indices = string.ascii_letters[nout:nout + ninp]
    gvar_indices = string.ascii_letters[nout + ninp]

    # make summation formula
    jac_indices = head_indices + out_indices + in_indices
    in_jac_indices = head_indices + in_indices + gvar_indices
    out_indices = head_indices + out_indices + gvar_indices
    formula = f'{jac_indices},{in_jac_indices}->{out_indices}'

    def gvar_function(x):

        # unpack the gvars
        in_mean = gvar.mean(x)
        in_jac, indices = jacobian(x)

        # apply function
        out_mean = func(in_mean)
        jac = deriv(in_mean)

        # check shapes match
        head_ndim = jac.ndim - nout - ninp
        assert jac.shape[:head_ndim] == in_jac.shape[:in_jac.ndim - 1 - ninp]

        # contract
        out_jac = jnp.einsum(formula, jac, in_jac)

        # pack output
        return from_jacobian(out_mean, out_jac, indices)

    @functools.wraps(func)
    def decorated_func(x):
        if isinstance(x, gvar.GVar):
            out = gvar_function(x)
            if not out.ndim:
                out = out.item()
            return out
        elif getattr(x, 'dtype', None) == object:
            return gvar_function(x)
        else:
            return func(x)

    return decorated_func

    # TODO add more than one argument or output. Possibly without taking
    # derivatives when it's not a gvar, i.e., merge the wrappers and cycle over
    # args. Also implement excluded => note that jnp.vectorize only supports
    # positional arguments, excluded takes in only indices, not names

@contextlib.contextmanager
def switchgvar():
    """
    Context manager to keep new gvars in a separate pool.

    Creating new primary gvars fills up memory permanently. This context manager
    keeps the gvars created within its context in a separate pool that is freed
    when all such gvars are deleted. They can not be mixed in operations with
    other gvars created outside of the context.
    
    Returns
    -------
    gvar : gvar.GVarFactory
        The new gvar-creating function that uses a new pool. The change is also
        reflected in the global `gvar.gvar`.

    See also
    --------
    gvar.switch_gvar, gvar.restore_gvar

    Examples
    --------

    >>> x = gvar.gvar(0, 1)
    >>> with lgp.switchgvar():
    >>>     y = gvar.gvar(0, 1)
    >>>     z = gvar.gvar(0, 1)
    >>> w = gvar.gvar(0, 1)
    >>> q = y + z  # allowed, y and z created in the same pool
    >>> p = x + w  # allowed, x and w created in the same pool
    >>> h = x + y  # x and y created in different pools: this will silently
    ...            # fail and possibly crash python immediately or later on

    """
    try:
        yield gvar.switch_gvar()
    finally:
        gvar.restore_gvar()



# lsqfitgp/_kernels/_zeta.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools

from jax import numpy as jnp

from .. import _special
from .. import _jaxext
from .. import _Kernel

__all__ = ['Zeta']

def check_nu(nu):
    with _jaxext.skipifabstract():
        assert 0 <= nu < jnp.inf, nu
        
def zeta_derivable(*, nu):
    check_nu(nu)
    with _jaxext.skipifabstract():
        return int(max(0, jnp.ceil(nu) - 1))

@_Kernel.crosskernel(bases=(_Kernel.AffineSpan, _Kernel.StationaryKernel,), maxdim=1, derivable=zeta_derivable)
def Zeta(delta, *, nu, **_):
    r"""
    
    Zeta kernel.
    
    .. math::
        k(\Delta)
        &= \frac{\Re F(\Delta, s)}{\zeta(s)} =
        \qquad (s = 1 + 2 \nu, \quad \nu \ge 0) \\
        &= \frac1{\zeta(s)} \sum_{k=1}^\infty
        \frac {\cos(2\pi k\Delta)} {k^s} = \\
        &= -(-1)^{s/2}
        \frac {(2\pi)^s} {2s!}
        \frac {\tilde B_s(\Delta)} {\zeta(s)}
        \quad \text{for even integer $s$.}
    
    It is equivalent to fitting with a Fourier series of period 1 with
    independent priors on the coefficients with mean zero and variance
    :math:`1/(\zeta(s)k^s)` for the :math:`k`-th term. Analogously to
    :class:`Matern`, the process is :math:`\lceil\nu\rceil - 1` times
    derivable, and the highest derivative is continuous iff :math:`\nu\bmod 1
    \ge 1/2`.
    
    The :math:`k = 0` term is not included in the summation, so the mean of the
    process over one period is forced to be zero.

    Reference: Petrillo (2022).
    
    """
    check_nu(nu)
    s = 1 + 2 * nu
    nupos = _special.periodic_zeta(delta, s) / _special.zeta(s)
    nuzero = jnp.where(delta % 1, 0, 1)
    return jnp.where(s > 1, nupos, nuzero)
    
    # return -(-1) ** (s // 2) * _special.scaled_periodic_bernoulli(s, delta) / jspecial.zeta(s, 1)
    
    # TODO use the bernoully version for integer even s, based on the type of
    # the input such that it's static, because it is much more accurate

    # TODO ND version. The separable product is not equivalent I think.
    
    # TODO the derivative w.r.t. nu is probably broken
    
@_Kernel.kernel(maxdim=1, derivable=False)
def ZetaFourier(k, q, *, nu, lloc, rloc, lscale, rscale, offset, ampl):
    check_nu(nu)
    s = 1 + 2 * nu
    lorder = jnp.ceil(k / 2)
    rorder = jnp.ceil(q / 2)
    lodd = k % 2
    rodd = q % 2
    var = ampl / (lorder ** s * _special.zeta(s))
    arg = 2 * jnp.pi * lorder * (lloc / lscale - rloc / rscale)
    return jnp.where(lorder == rorder,
        jnp.where(lodd == rodd,
            jnp.where(lorder, var * jnp.cos(arg), offset),
            var * jnp.sin(arg) * jnp.where(lodd, 1, -1),
        ),
        0,
    )

def crosszeta_derivable(*, nu, **_):
    return 0, zeta_derivable(nu=nu)

@_Kernel.crosskernel(bases=(_Kernel.PreservedBySwap, _Kernel.CrossKernel), maxdim=1, derivable=crosszeta_derivable)
def CrossZetaFourier(k, y, *, nu, lloc, rloc, lscale, rscale, offset, ampl):
    check_nu(nu)
    s = 1 + 2 * nu
    order = jnp.ceil(k / 2)
    odd = k % 2
    var = ampl / (order ** s * _special.zeta(s))
    arg = 2 * jnp.pi * order * (lloc / lscale + (y - rloc) / rscale)
    return jnp.where(odd,
        var * jnp.sin(arg),
        jnp.where(order, var * jnp.cos(arg), offset),
    )

fourier_doc = r"""

Compute the Fourier series transform of the function.

.. math::

    T(f)(k) = \begin{cases}
            \frac2T \int_0^T \mathrm dx\, f(x)
            \cos\left(\frac{2\pi}T \frac k2 x\right)
            & \text{if $k$ is even} \\
            \frac2T \int_0^T \mathrm dx\, f(x)
            \sin\left(\frac{2\pi}T \frac{k+1}2 x\right)
            & \text{if $k$ is odd}
        \end{cases}
    
The period :math:`T` is 1.

"""

def fourier_argparser(do):
    return do if do else None

def translkw(*, dynkw, **initkw):
    return dict(**dynkw, **initkw)

Zeta.make_linop_family('fourier', ZetaFourier, CrossZetaFourier, translkw=translkw, doc=fourier_doc, argparser=fourier_argparser)

# TODO
# - test the transf with rescalings (what cross check can I do?)
# - track affine transf in CrossZetaFourier too
# - make Zeta support non-sym affine ops (I think I need to define CrossZeta
#    then subclass to Zeta(CrossZeta, Kernel)
# - consider renaming fourier to fourier_series when I rewrite transf system



# lsqfitgp/_Kernel/_ops.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" register linops on CrossKernel and AffineSpan """

import functools
import numbers
import sys

from jax import numpy as jnp
import numpy

from .. import _jaxext
from .. import _Deriv
from .. import _array

from . import _util
from ._crosskernel import CrossKernel, AffineSpan

def rescale_argparser(fun):
    if not callable(fun):
        raise ValueError("argument to 'rescale' must be a function")
    return fun

@functools.partial(CrossKernel.register_corelinop, argparser=rescale_argparser)
def rescale(core, xfun, yfun):
    r"""
    
    Rescale the output of the function.
    
    .. math::
        T(f)(x) = \mathrm{fun}(x) f(x)
    
    Parameters
    ----------
    xfun, yfun : callable or None
        Functions from the type of the arguments of the kernel to scalar.
    
    """
    if not xfun:
        return lambda x, y, **kw: core(x, y, **kw) * yfun(y)
    elif not yfun:
        return lambda x, y, **kw: xfun(x) * core(x, y, **kw)
    else:
        return lambda x, y, **kw: xfun(x) * core(x, y, **kw) * yfun(y)

@CrossKernel.register_xtransf
def derivable(derivable):
    """
    Specify the degree of derivability of the function.

    Parameters
    ----------
    xderivable, yderivable: int or None
        Degree of derivability of the function. None means unknown.

    Notes
    -----
    The derivability check is hardcoded into the kernel core and it is not
    possible to remove it afterwards by applying ``'derivable'`` again with a
    higher limit.

    """
    if isinstance(derivable, bool):
        derivable = sys.maxsize if derivable else 0
    elif not isinstance(derivable, numbers.Integral) or derivable < 0:
        raise ValueError(f'derivability degree {derivable!r} not valid')

    def error_func(current, n):
        raise ValueError(f'Took {current} derivatives > limit {n} on argument '
            'of a kernel. This error may be spurious if there are '
            'derivatives on values that define the input to the kernel, for '
            'example if a hyperparameter enters the calculation of x. To '
            'suppress the error, initialize the kernel with derivable=True.')

    def xtransf(x):
        if hasattr(x, 'dtype'):
            # this branch handles limit_derivatives not accepting non-jax types
            # because of being based on jax.custom_jvp; this restriction on
            # custom_jvp appeared in jax 0.4.17
            if x.dtype.names is not None:
                x = _array.StructuredArray(x) # structured arrays are not
                    # compatible with jax but common in lsqfitgp, so I wrap them
            elif not _jaxext.is_jax_type(x.dtype):
                return x # since anyway there would be an error if a derivative
                    # tried to pass through a non-jax type
        return _jaxext.limit_derivatives(x, n=derivable, error_func=error_func)

    return xtransf

    # TODO this system does not ignore additional derivatives that are not
    # taken by .transf('diff'). Plan:
    # - understand how to associate a jax transformation to a frame
    # - make a context manager with a global stack
    # - at initialization it takes the frames of the derivatives made by diff
    # - the context is around calling core in diff.newcore
    # - it adds the frames to the stack
    # - derivable asks the context manager class to find the frames in x
    # - raises if their number is above the limit

def _asfloat(x):
    return x.astype(_jaxext.float_type(x))

def diff_argparser(deriv):
    deriv = _Deriv.Deriv(deriv)
    return deriv if deriv else None

@functools.partial(CrossKernel.register_corelinop, argparser=diff_argparser)
def diff(core, xderiv, yderiv):
    r"""
    
    Derive the function.
    
    .. math::
        T(f)(x) = \frac{\partial^n f}{\partial x^n} (x)
    
    Parameters
    ----------
    xderiv, yderiv : Deriv_like
        A `Deriv` or something that can be converted to a `Deriv`.
    
    Raises
    ------
    RuntimeError
        The derivative orders are greater than the `derivative` attribute.
        
    """

    # reparse derivatives because they could be None
    xderiv = _Deriv.Deriv(xderiv)
    yderiv = _Deriv.Deriv(yderiv)

    # wrapper of kernel with derivable arguments unpacked
    def f(x, y, *args, **kw):
        i = -1
        if not xderiv.implicit:
            for i, dim in enumerate(xderiv):
                x = x.at[dim].set(args[i])
        if not yderiv.implicit:
            for j, dim in enumerate(yderiv):
                y = y.at[dim].set(args[1 + i + j])
        return core(x, y, **kw)
    
    # last x index in case iteration on x does not run but does on y
    i = -1

    # derive w.r.t. first argument
    if xderiv.implicit:
        for _ in range(xderiv.order):
            f = _jaxext.elementwise_grad(f, 0)
    else:
        for i, dim in enumerate(xderiv):
            for _ in range(xderiv[dim]):
                f = _jaxext.elementwise_grad(f, 2 + i)

    # derive w.r.t. second argument
    if yderiv.implicit:
        for _ in range(yderiv.order):
            f = _jaxext.elementwise_grad(f, 1)
    else:
        for j, dim in enumerate(yderiv):
            for _ in range(yderiv[dim]):
                f = _jaxext.elementwise_grad(f, 2 + 1 + i + j)

    # check derivatives are ok for actual input arrays, wrap structured arrays
    def process_arg(x, deriv, pos):
        if x.dtype.names is not None:
            for dim in deriv:
                if dim not in x.dtype.names:
                    raise ValueError(f'derivative along missing field {dim!r} '
                        f'on {pos} argument')
                if not jnp.issubdtype(x.dtype[dim], jnp.number):
                    raise TypeError(f'derivative along non-numeric field '
                        f'{dim!r} on {pos} argument')
            return _array.StructuredArray(x)
        elif not deriv.implicit:
            raise ValueError('derivative on named fields with non-structured '
                f'array on {pos} argument')
        elif not jnp.issubdtype(x.dtype, jnp.number):
            raise TypeError(f'derivative along non-numeric array on '
                f'{pos} argument')
        return x
    
    def newcore(x, y, **kw):
        x = process_arg(x, xderiv, 'left')
        y = process_arg(y, yderiv, 'right')
                            
        args = []

        if not xderiv.implicit:
            for dim in xderiv:
                args.append(_asfloat(x[dim]))
        elif xderiv:
            x = _asfloat(x)
        
        if not yderiv.implicit:
            for dim in yderiv:
                args.append(_asfloat(y[dim]))
        elif yderiv:
            y = _asfloat(y)

        return f(x, y, *args, **kw)

    return newcore

@CrossKernel.register_xtransf
def xtransf(fun):
    r"""
    
    Transform the inputs of the function.
    
    .. math::
        T(f)(x) = f(\mathrm{fun}(x))
    
    Parameters
    ----------
    xfun, yfun : callable or None
        Functions mapping a new kind of input to the kind of input accepted by
        the kernel.
    
    """
    if not callable(fun):
        raise ValueError("argument to 'xtransf' must be a function")
    return fun

@CrossKernel.register_xtransf
def dim(dim):
    """
    Restrict the function to a field of a structured input::

        T(f)(x) = f(x[dim])

    If the array is not structured, an exception is raised. If the field for
    name `dim` has a nontrivial shape, the array passed to the kernel is still
    structured but has only field `dim`.

    Parameters
    ----------
    xdim, ydim: None, str, list of str
        Field names or lists of field names.

    """
    if not isinstance(dim, (str, list)):
        raise TypeError(f'dim must be a (list of) string, found {dim!r}')
    def fun(x):
        if x.dtype.names is None:
            raise ValueError(f'cannot get dim={dim!r} from non-structured input')
        elif x.dtype[dim].shape:
            return x[[dim]]
        else:
            return x[dim]
    return fun

@CrossKernel.register_xtransf
def maxdim(maxdim):
    """

    Restrict the process to a maximum input dimensionality.

    Parameters
    ----------
    xmaxdim, ymaxdim: None, int
        Maximum dimensionality of the input.
    
    Notes
    -----
    Once applied a restriction, the check is hardcoded into the kernel core and
    it is not possible to remove it by applying again `maxdim` with a larger
    limit.

    """
    if not isinstance(maxdim, numbers.Integral) or maxdim < 0:
        raise ValueError(f'maximum dimensionality {maxdim!r} not valid')

    def fun(x):
        nd = _array._nd(x.dtype)
        with _jaxext.skipifabstract():
            if nd > maxdim:
                raise ValueError(f'kernel applied to input with {nd} '
                    f'fields > maxdim={maxdim}')
        return x
    
    return fun

@CrossKernel.register_xtransf
def loc(loc):
    r"""
    Translate the process inputs:

    .. math::
        T(f)(x) = f(x - \mathrm{loc})

    Parameters
    ----------
    xloc, yloc: None, number
        Translations.

    """
    with _jaxext.skipifabstract():
        assert -jnp.inf < loc < jnp.inf, loc
    return lambda x: _util.ufunc_recurse_dtype(lambda x: x - loc, x)

@CrossKernel.register_xtransf
def scale(scale):
    r"""
    Rescale the process inputs:

    .. math::
        T(f)(x) = f(x / \mathrm{scale})

    Parameters
    ----------
    xscale, yscale: None, number
        Rescaling factors.

    """
    with _jaxext.skipifabstract():
        assert 0 < scale < jnp.inf, scale
    return lambda x: _util.ufunc_recurse_dtype(lambda x: x / scale, x)

def normalize_argparser(do):
    return do if do else None

@functools.partial(CrossKernel.register_corelinop, argparser=normalize_argparser)
def normalize(core, dox, doy):
    r"""
    Rescale the process to unit variance.

    .. math::
        T(f)(x) &= f(x) / \sqrt{\mathrm{Std}[f(x)]} \\
                &= f(x) / \sqrt{\mathrm{kernel}(x, x)}

    Parameters
    ----------
    dox, doy : bool
        Whether to rescale.
    """
    if dox and doy:
        return lambda x, y, **kw: core(x, y, **kw) / jnp.sqrt(core(x, x, **kw) * core(y, y, **kw))
    elif dox:
        return lambda x, y, **kw: core(x, y, **kw) / jnp.sqrt(core(x, x, **kw))
    else:
        return lambda x, y, **kw: core(x, y, **kw) / jnp.sqrt(core(y, y, **kw))

@CrossKernel.register_corelinop
def cond(core, cond1, cond2, other):
    r"""

    Switch between two independent processes based on a condition.

    .. math::
        T(f, g)(x) = \begin{cases}
            f(x) & \text{if $\mathrm{cond}(x)$,} \\
            g(x) & \text{otherwise.}
        \end{cases}
        
    Parameters
    ----------
    cond1, cond2 : callable
        Function that is applied on an array of points and must return
        a boolean array with the same shape.
    other :
        Kernel of the process used where the condition is false.
    
    """
    def newcore(x, y, **kw):        
        xcond = cond1(x)
        ycond = cond2(y)
        r = jnp.where(xcond & ycond, core(x, y, **kw), other(x, y, **kw))
        return jnp.where(xcond ^ ycond, 0, r)
    
    return newcore

    # TODO add a function `choose` to extend `cond`,
    # kernel0.linop('choose', kernel1, kernel2, ..., lambda x: x['index'])

AffineSpan.inherit_transf('maxdim')
AffineSpan.inherit_transf('derivable')

@functools.partial(AffineSpan.register_linop, transfname='loc')
def affine_loc(tcls, self, xloc, yloc):
    dynkw = dict(self.dynkw)
    newself = tcls.super_transf('loc', self, xloc, yloc)
    dynkw['lloc'] = dynkw['lloc'] + xloc * dynkw['lscale']
    dynkw['rloc'] = dynkw['rloc'] + yloc * dynkw['rscale']
    return newself._clone(self.__class__, dynkw=dynkw)

@functools.partial(AffineSpan.register_linop, transfname='scale')
def affine_scale(tcls, self, xscale, yscale):
    dynkw = dict(self.dynkw)
    newself = tcls.super_transf('scale', self, xscale, yscale)
    dynkw['lscale'] = dynkw['lscale'] * xscale
    dynkw['rscale'] = dynkw['rscale'] * yscale
    return newself._clone(self.__class__, dynkw=dynkw)



# lsqfitgp/_special/_taylor.py
#
# Copyright (c) 2022, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools

import jax
from jax import numpy as jnp
from jax.scipy import special as jspecial

@functools.partial(jax.custom_jvp, nondiff_argnums=(0, 1, 2, 3))
def taylor(coefgen, args, n, m, x):
    """
    coefgen : function = start, end -> taylor coefficients for powers start:end
    args : tuple = additional arguments to coefgen
    n : int = derivation order
    m : int = number of coefficients used
    x : argument
    """
    c = coefgen(n, n + m, *args)
    k = jnp.arange(n, n + m)
    c = c * jnp.exp(jspecial.gammaln(1 + k) - jspecial.gammaln(1 + k - n))
    return jnp.polyval(c[::-1], x)

@taylor.defjvp
def taylor_jvp(coefgen, args, n, m, primals, tangents):
    x, = primals
    xt, = tangents
    return taylor(coefgen, args, n, m, x), taylor(coefgen, args, n + 1, m, x) * xt



# lsqfitgp/copula/_base.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" define DistrBase """

import abc
import functools
import collections

import gvar
import numpy

class DistrBase(metaclass=abc.ABCMeta):
    r"""

    Abstract base class to represent (trees of) probability distributions.

    Attributes
    ----------
    in_shape : tuple of int
        The core shape of the input array to `partial_invfcn`.
    shape : (tree of) tuple of int
        The core shape of the output array of `partial_invfcn`.
    dtype : (tree of) dtype
        The dtype of the output array of `partial_invfcn`.
    distrshape : (tree of) tuple of int
        The sub-core shape of the output array of `partial_invfcn` that
        represents the atomic shape of the distribution.

    Methods
    -------
    partial_invfcn :
        Transformation function from a (multivariate) Normal variable to the
        target random variable.
    add_distribution :
        Register the distribution for usage with `gvar.BufferDict`.
    gvars :
        Return an array of gvars with the appropriate shape for usage with
        `gvar.BufferDict`.

    See also
    --------
    Distr, Copula

    """

    def __init_subclass__(cls, **kw):
        super().__init_subclass__(**kw)
        cls._named = {}

    in_shape = NotImplemented
    shape = NotImplemented
    dtype = NotImplemented
    distrshape = NotImplemented

    def partial_invfcn(self, x):
        """
            
        Map independent Normal variables to the desired distribution.

        This function is a generalized ufunc. It is jax traceable and
        differentiable one time. It supports arrays of gvars as input. The
        attributes `in_shape` and `shape` give the core shapes.

        Parameters
        ----------
        x : ``(..., *in_shape)`` array
            An array of values representing draws of i.i.d. Normal variates.

        Returns
        -------
        y : (tree of) ``(..., *shape)`` array
            An array of values representing draws of the desired distribution.

        """
        return self._partial_invfcn(x)

    @abc.abstractmethod
    def _partial_invfcn(self, x):
        pass
    
    def _is_same_family(self, invfcn):
        return getattr(invfcn, '__self__', None).__class__ is self.__class__

    def add_distribution(self, name):
        """

        Register the distribution for usage with `gvar.BufferDict`.

        Parameters
        ----------
        name : str
            The name to use for the distribution. It must be globally unique,
            and it should not contain parentheses. To redefine a distribution
            with the same name, use `gvar.BufferDict.del_distribution` first.
            However, it is allowed to reuse the name if the distribution family,
            shape and parameters are identical to those used for the existing
            definition.

        See also
        --------
        gvar.BufferDict.add_distribution, gvar.BufferDict.del_distribution

        """

        if gvar.BufferDict.has_distribution(name):
            invfcn = gvar.BufferDict.invfcn[name]
            if not self._is_same_family(invfcn):
                raise ValueError(f'distribution {name} already defined')
            existing = self._named[name]
            if existing != self._staticdescr:
                raise ValueError('Attempt to overwrite existing'
                    f' {self.__class__.__name__} distribution with name {name}')
                # cls._named is not updated by
                # gvar.BufferDict.del_distribution, but it is not a problem
        
        else:
            gvar.BufferDict.add_distribution(name, self.partial_invfcn)
            self._named[name] = self._staticdescr
    
    def gvars(self):
        """

        Return an array of gvars intended as value in a `gvar.BufferDict`.

        Returns
        -------
        gvars : array of gvars
            An array of i.i.d. standard Normal primary gvars with shape
            `in_shape`.

        """

        return gvar.gvar(numpy.zeros(self.in_shape), numpy.ones(self.in_shape))

    @abc.abstractmethod
    def __repr__(self, path='', cache=None):
        """ produce a representation where no object appears more than once,
        later appearances are replaced by a user-friendly identifier """
        if cache is None:
            cache = {}
        if self in cache:
            return cache[self]
        cache[self] = f'<{path}>'
        return cache

    class _Path(collections.namedtuple('Path', ['path'])): pass

    @abc.abstractmethod
    def _compute_staticdescr(self, path, cache):
        """ compute static description of self, can be compared """
        if self in cache:
            return cache[self]
        cache[self] = self._Path(path)

    @functools.cached_property
    def _staticdescr(self):
        return self._compute_staticdescr([], {})

    @abc.abstractmethod
    def _compute_in_size(self, cache):
        """ compute input size to partial_invfcn, without double counting """
        if self in cache:
            return 0
        cache.add(self)

    @abc.abstractmethod
    def _partial_invfcn_internal(self, x, i, cache):
        assert x.ndim == 1
        if self in cache:
            return cache[self], i



# lsqfitgp/_special/_bessel.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools

from scipy import special
import jax
from jax import numpy as jnp
from jax.scipy import special as jspecial

from .. import _jaxext
from . import _gamma

j0 = _jaxext.makejaxufunc(special.j0, lambda x: -j1(x))
j1 = _jaxext.makejaxufunc(special.j1, lambda x: (j0(x) - jv(2, x)) / 2.0)
jv = _jaxext.makejaxufunc(special.jv, None, lambda v, z: jvp(v, z, 1))
jvp = _jaxext.makejaxufunc(special.jvp, None, lambda v, z, n: jvp(v, z, n + 1), None, excluded=(2,))

kv = _jaxext.makejaxufunc(special.kv, None, lambda v, z: kvp(v, z, 1))
kvp = _jaxext.makejaxufunc(special.kvp, None, lambda v, z, n: kvp(v, z, n + 1), None, excluded=(2,))

iv = _jaxext.makejaxufunc(special.iv, None, lambda v, z: ivp(v, z, 1))
ivp = _jaxext.makejaxufunc(special.ivp, None, lambda v, z, n: ivp(v, z, n + 1), None, excluded=(2,))

# See jax #1870, #2466, #9956, #11002 and
# https://github.com/josipd/jax/blob/master/jax/experimental/jambax.py
# to implement special functions in jax with numba

@functools.partial(jax.custom_jvp, nondiff_argnums=(0,))
@jax.jit
def jvmodx2(nu, x2):
    x = jnp.sqrt(x2)
    normal = (x / 2) ** -nu * jv(nu, x)
    return jnp.where(x2, normal, 1 / _gamma.gamma(nu + 1))

# (1/x d/dx)^m (x^-v J_v(x)) = (-1)^m x^-(v+m) J_v+m(x)
#                                   (Abramowitz and Stegun, p. 361, 9.1.30)
# --> 1/x d/dx (x^-v J_v(x)) = -x^-(v+1) J_v+1(x)
# --> d/dx (x^-v J_v(x)) = -x^-v J_v+1(x)
# --> d/ds ~J_v(s) =
#   = d/ds (√s/2)^-v J_v(√s) =
#   = 2^v d/ds √s^-v J_v(√s) =
#   = -2^v √s^-v J_v+1(√s) 1/2√s =
#   = -2^(v-1) √s^-(v+1) J_v+1(√s) =
#   = -1/4 (√s/2)^(v+1) J_v+1(√s) =
#   = -1/4 ~J_v+1(s)

@jvmodx2.defjvp
def jvmodx2_jvp(nu, primals, tangents):
    x2, = primals
    x2t, = tangents
    return jvmodx2(nu, x2), -x2t * jvmodx2(nu + 1, x2) / 4

@functools.partial(jax.custom_jvp, nondiff_argnums=(0, 2))
@functools.partial(jax.jit, static_argnums=(2,))
def kvmodx2(nu, x2, norm_offset=0):
    x = jnp.sqrt(x2)
    normal = 2 / _gamma.gamma(nu + norm_offset) * (x / 2) ** nu * kv(nu, x)
    atzero = 1 / jnp.prod(nu + jnp.arange(norm_offset))
    atzero = jnp.where(nu > 0, atzero, 1) # for nu < 0 the correct limit
                                          # would be inf, but in practice it
                                          # gets cancelled by a stronger 0
                                          # when taking derivatives of Matern
                                          # and this is a cheap way to avoid
                                          # nans
    return jnp.where(x2, normal, atzero)

# d/dx (x^v Kv(x)) = -x^v Kv-1(x)       (Abrahamsen 1997, p. 43)
# d/ds ~Kv(s) =
#   = d/ds (√s/2)^v Kv(√s) =
#   = 2^-v d/ds (√s)^v Kv(√s) =
#   = -2^-v (√s)^v Kv-1(√s) 1/(2√s) =
#   = -2^-(v+1) √s^(v-1) Kv-1(√s) =
#   = -1/4 (√s/2)^(v-1) Kv-1(√s) =
#   = -1/4 ~Kv-1(s)

@kvmodx2.defjvp
def kvmodx2_jvp(nu, norm_offset, primals, tangents):
    x2, = primals
    x2t, = tangents
    primal = kvmodx2(nu, x2, norm_offset)
    tangent = -x2t * kvmodx2(nu - 1, x2, norm_offset + 1) / 4
    return primal, tangent

@functools.partial(jax.custom_jvp, nondiff_argnums=(1,))
@functools.partial(jax.jit, static_argnums=(1,))
def kvmodx2_hi(x2, p):
    # nu = p + 1/2, p integer >= 0
    x = jnp.sqrt(x2)
    poly = 1
    for k in reversed(range(p)):
        c_kp1_over_ck = (p - k) / ((2 * p - k) * (k + 1))
        poly = 1 + poly * c_kp1_over_ck * 2 * x
    return jnp.exp(-x) * poly

@kvmodx2_hi.defjvp
def kvmodx2_hi_jvp(p, primals, tangents):
    x2, = primals
    x2t, = tangents
    primal = kvmodx2_hi(x2, p)
    if p == 0:
        x = jnp.sqrt(x2)
        tangent = -x2t * jnp.exp(-x) / (2 * x) # <--- problems!
    else:
        tangent = -x2t / (p - 1/2) * kvmodx2_hi(x2, p - 1) / 4
    return primal, tangent



# lsqfitgp/_GP/_elements.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import abc
import functools
import warnings
import math

import gvar
import numpy
from scipy import sparse
import jax
from jax import numpy as jnp

from .. import _Deriv
from .. import _array
from .. import _jaxext
from .. import _gvarext
from .. import _linalg

from . import _base

class GPElements(_base.GPBase):

    def __init__(self, *, checkpos, checksym, posepsfac, halfmatrix):
        self._elements = dict() # key -> _Element
        self._covblocks = dict() # (key, key) -> matrix (2d flattened)
        self._priordict = {} # key -> gvar array (shaped)
        self._checkpositive = bool(checkpos)
        self._posepsfac = float(posepsfac)
        self._checksym = bool(checksym)
        self._halfmatrix = bool(halfmatrix)
        self._dtype = None
        assert not (halfmatrix and checksym)

    def _clone(self):
        newself = super()._clone()
        newself._elements = self._elements.copy()
        newself._covblocks = self._covblocks.copy()
        newself._priordict = self._priordict.copy()
        newself._checkpositive = self._checkpositive
        newself._posepsfac = self._posepsfac
        newself._checksym = self._checksym
        newself._halfmatrix = self._halfmatrix
        newself._dtype = self._dtype
        return newself

    @staticmethod
    def _concatenate(alist):
        """
        Decides to use numpy.concatenate or jnp.concatenate depending on the
        input to support gvars.
        """
        if any(a.dtype == object for a in alist):
            return numpy.concatenate(alist)
        else:
            return jnp.concatenate(alist)

    @staticmethod
    def _triu_indices_and_back(n):
        """
        Return indices to get the upper triangular part of a matrix, and indices
        to convert a flat array of upper triangular elements to a symmetric
        matrix.
        """
        ix, iy = jnp.triu_indices(n)
        q = jnp.empty((n, n), ix.dtype)
        a = jnp.arange(ix.size)
        q = q.at[ix, iy].set(a)
        q = q.at[iy, ix].set(a)
        return ix, iy, q

    class _Element(abc.ABC):
        """
        Abstract class for an object holding information associated to a key in
        a GP object.
        """
    
        @property
        @abc.abstractmethod
        def shape(self): # pragma: no cover
            """Output shape"""
            pass
    
        @property
        def size(self):
            return math.prod(self.shape)

    class _Points(_Element):
        """Points where the process is evaluated"""
    
        def __init__(self, x, deriv, proc):
            assert isinstance(x, (numpy.ndarray, jnp.ndarray, _array.StructuredArray))
            assert isinstance(deriv, _Deriv.Deriv)
            self.x = x
            self.deriv = deriv
            self.proc = proc
    
        @property
        def shape(self):
            return self.x.shape

    class _LinTransf(_Element):
        """Linear transformation of other _Element objects"""
    
        shape = None
    
        def __init__(self, transf, keys, shape):
            self.transf = transf
            self.keys = keys
            self.shape = shape
    
        def matrices(self, gp):
            """
            Matrix coefficients of the transformation (with flattened inputs
            and output)
            """
            elems = [gp._elements[key] for key in self.keys]
            matrices = []
            transf = jax.vmap(self.transf, 0, 0)
            for i, elem in enumerate(elems):
                inputs = [
                    jnp.eye(elem.size).reshape((elem.size,) + elem.shape)
                    if j == i else
                    jnp.zeros((elem.size,) + ej.shape)
                    for j, ej in enumerate(elems)
                ]
                output = transf(*inputs).reshape(elem.size, self.size).T
                matrices.append(output)
            return matrices

    class _Cov(_Element):
        """User-provided covariance matrix block(s)"""
    
        shape = None
    
        def __init__(self, blocks, shape):
            """ blocks = dict (key, key) -> matrix """
            self.blocks = blocks
            self.shape = shape

    @_base.newself
    def addx(self, x, key=None, *, deriv=0, proc=_base.GPBase.DefaultProcess):
        """
        
        Add points where the Gaussian process is evaluated.
        
        The GP object keeps the various x arrays in a dictionary. If ``x`` is an
        array, you have to specify its dictionary key with the ``key`` parameter.
        Otherwise, you can directly pass a dictionary for ``x``.
        
        To specify that on the given ``x`` a derivative of the process instead of
        the process itself should be evaluated, use the parameter ``deriv``.
        
        `addx` may or may not copy the input arrays.
        
        Parameters
        ----------
        x : array or dictionary of arrays
            The points to be added.
        key : hashable
            If ``x`` is an array, the dictionary key under which ``x`` is added.
            Can not be specified if ``x`` is a dictionary.
        deriv : Deriv-like
            Derivative specification. A `Deriv` object or something that
            can be converted to `Deriv`.
        proc : hashable
            The process to be evaluated on the points. If not specified, use
            the default process.
        
        """
        
        # TODO after I implement block solving, add per-key covariance matrix
        # flags.
        
        # TODO add `copy` parameter, default False, to copy the input arrays
        # if they are numpy arrays.
        
        # this interface does not allow adding a single dictionary as x element
        # unless it's wrapped as a 0d numpy array, but this is for the best
        
        deriv = _Deriv.Deriv(deriv)

        if proc not in self._procs:
            raise KeyError(f'process named {proc!r} not found')
                
        if hasattr(x, 'keys'):
            if key is not None:
                raise ValueError('can not specify key if x is a dictionary')
            if None in x:
                raise ValueError('None key in x not allowed')
        else:
            if key is None:
                raise ValueError('x is not dictionary but key is None')
            x = {key: x}
        
        for key in x:
            if key in self._elements:
                raise KeyError('key {!r} already in GP'.format(key))
            
            gx = x[key]
            
            # Convert to JAX array, numpy array or StructuredArray.
            # convert eagerly to jax to avoid problems with tracing.
            gx = _array._asarray_jaxifpossible(gx)

            # Check dtype is compatible with previous arrays.
            # TODO since we never concatenate arrays we could allow a less
            # strict compatibility. In principle we could allow really anything
            # as long as the kernel eats it, but this probably would let bugs
            # through without being really ever useful. What would make sense
            # is checking the dtype structure matches recursively and check
            # concrete dtypes of fields can be casted.
            # TODO result_type is too lax. Examples: str, float -> str,
            # object, float -> object. I should use something like the
            # ordering function in updowncast.py.
            if self._dtype is not None:
                try:
                    self._dtype = numpy.result_type(self._dtype, gx.dtype)
                    # do not use jnp.result_type, it does not support
                    # structured types
                except TypeError:
                    msg = 'x[{!r}].dtype = {!r} not compatible with {!r}'
                    msg = msg.format(key, gx.dtype, self._dtype)
                    raise TypeError(msg)
            else:
                self._dtype = gx.dtype

            # Check that the derivative specifications are compatible with the
            # array data type.
            if gx.dtype.names is None:
                if not deriv.implicit:
                    raise ValueError('x has no fields but derivative has')
            else:
                for dim in deriv:
                    if dim not in gx.dtype.names:
                        raise ValueError(f'deriv field {dim!r} not in x')
            
            self._elements[key] = self._Points(gx, deriv, proc)

    def _get_x_dtype(self):
        """ Get the data type of x points """
        return self._dtype
        
    def addtransf(self, tensors, key, *, axes=1):
        """
        
        Apply a linear transformation to already specified process points. The
        result of the transformation is represented by a new key.
        
        Parameters
        ----------
        tensors : dict
            Dictionary mapping keys of the GP to arrays/scalars. Each array is
            matrix-multiplied with the process array represented by its key,
            while scalars are just multiplied. Finally, the keys are summed
            over.
        key : hashable
            A new key under which the transformation is placed.
        axes : int
            Number of axes to be summed over for matrix multiplication,
            referring to trailing axes for tensors in ` tensors``, and to
            heading axes for process points. Default 1.
        
        Returns
        -------
        gp : GP
            A new GP object with the applied modifications.
       
        Notes
        -----
        The multiplication between the tensors and the process is done with
        np.tensordot with, by default, 1-axis contraction. For >2d arrays this
        is different from numpy's matrix multiplication, which would act on the
        second-to-last dimension of the second array.
        
        """        
        # Note: it may seem nice that when an array has less axes than `axes`,
        # the summation would be restricted only on the existing axes. However
        # this brings about the ambiguous case where only one of the factors has
        # not enough axes. How many axes do you sum over on the other?
        
        # Check axes.
        assert isinstance(axes, int) and axes >= 0, axes
        
        # Check key.
        if key is None:
            raise ValueError('key can not be None')
        if key in self._elements:
            raise KeyError(f'key {key!r} already in GP')
        
        # Check keys.
        for k in tensors:
            if k not in self._elements:
                raise KeyError(k)
        
        # Check tensors and convert them to jax arrays.
        if len(tensors) == 0:
            raise ValueError('empty tensors, undetermined output shape')
        tens = {}
        for k, t in tensors.items():
            t = jnp.asarray(t)
            # no need to check dtype since jax supports only numerical arrays
            with _jaxext.skipifabstract():
                if self._checkfinite and not jnp.all(jnp.isfinite(t)):
                    raise ValueError(f'tensors[{k!r}] contains infs/nans')
            rshape = self._elements[k].shape
            if t.shape and t.shape[t.ndim - axes:] != rshape[:axes]:
                raise ValueError(f'tensors[{k!r}].shape = {t.shape!r} can not be multiplied with shape {rshape!r} with {axes}-axes contraction')
            tens[k] = t
        
        # Check shapes broadcast correctly.
        arrays = tens.values()
        elements = (self._elements[k] for k in tens)
        shapes = (
            t.shape[:t.ndim - axes] + e.shape[axes:] if t.shape else e.shape
            for t, e in zip(arrays, elements)
        )
        try:
            shape = jnp.broadcast_shapes(*shapes)
        except ValueError:
            msg = 'can not broadcast tensors with shapes ['
            msg += ', '.join(repr(t.shape) for t in arrays)
            msg += '] contracted with arrays with shapes ['
            msg += ', '.join(repr(e.shape) for e in elements) + ']'
            raise ValueError(msg)
        
        # Define linear transformation.
        def equiv_lintransf(*args):
            assert len(args) == len(tens)
            out = None
            for a, (k, t) in zip(args, tens.items()):
                if t.shape:
                    b = jnp.tensordot(t, a, axes)
                else:
                    b = t * a
                if out is None:
                    out = b
                else:
                    out = out + b
            return out
        keys = list(tens.keys())
        return self.addlintransf(equiv_lintransf, keys, key, checklin=False)
    
    @_base.newself
    def addlintransf(self, transf, keys, key, *, checklin=None):
        """
        
        Define a finite linear transformation of the evaluated process.
        
        Parameters
        ----------
        transf : callable
            A function with signature ``f(array1, array2, ...) -> array`` which
            computes the linear transformation. The function must be
            jax-traceable, i.e., use jax.numpy instead of numpy.
        keys : sequence
            Keys of parts of the process to be passed as inputs to the
            transformation.
        key : hashable
            The key of the newly defined points.
        checklin : bool
            If True (default), check that the given function is linear in its
            inputs. The default can be overridden at initialization of the GP
            object. Note that an affine function (x -> a + bx) is not linear.
        
        Raises
        ------
        RuntimeError :
            The transformation seems not to be linear. To disable the linearity
            check, initialize the GP with ``checklin=False``.
        
        """
        
        # TODO elementwise operations can be applied more efficiently to
        # primary gvars (tipical case), so the method could use an option
        # `elementwise`. What is the reliable way to check it is indeed
        # elementwise with a single random vector? Zero items of the tangent
        # at random with p=0.5 and check they stay zero? (And of course check
        # the shape is preserved.)
        
        # Check key.
        if key is None:
            raise ValueError('key can not be None')
        if key in self._elements:
            raise KeyError(f'key {key!r} already in GP')
        
        # Check keys.
        for k in keys:
            if k not in self._elements:
                raise KeyError(k)
        
        # Determine shape.
        class ArrayMockup:
            def __init__(self, elem):
                self.shape = elem.shape
                self.dtype = float
        inp = [ArrayMockup(self._elements[k]) for k in keys]
        out = jax.eval_shape(transf, *inp)
        shape = out.shape
        
        # Check that the transformation is linear.
        if checklin is None:
            checklin = self._checklin
        if checklin:
            shapes = [self._elements[k].shape for k in keys]
            self._checklinear(transf, shapes)
        
        self._elements[key] = self._LinTransf(transf, keys, shape)
    
    @_base.newself
    def addcov(self, covblocks, key=None, *, decomps=None):
        """
        
        Add user-defined prior covariance matrix blocks.
        
        Covariance matrices defined with `addcov` represent arbitrary
        finite-dimensional zero-mean Gaussian variables, assumed independent
        from all other variables in the GP object.
                
        Parameters
        ----------
        covblocks : array or dictionary of arrays
            If an array: a covariance matrix (or tensor) to be added under key
            ``key``. If a dictionary: a mapping from pairs of keys to the
            corresponding covariance matrix blocks. A missing off-diagonal
            block in the dictionary is interpreted as a matrix of zeros,
            unless the corresponding transposed block is specified.
        key : hashable
            If ``covblocks`` is an array, the dictionary key under which
            ``covblocks`` is added. Can not be specified if ``covblocks`` is a
            dictionary.
        decomps : Decomposition or dict of Decompositions
            Pre-computed decompositions of (not necessarily all) diagonal
            blocks, as produced by `decompose`. The keys are single
            GP keys and not pairs like in ``covblocks``.
        
        Raises
        ------
        KeyError :
            A key is already used in the GP.
        ValueError :
            ``covblocks`` and/or ``key`` and ``decomps`` are malformed or
            inconsistent.
        TypeError :
            Wrong type of ``covblocks`` or ``decomps``.
        
        """
        
        # TODO maybe allow passing only the lower/upper triangular part for
        # the diagonal blocks, like I meta-allow for out of diagonal blocks?
        
        # TODO with multiple blocks and a single decomp, the decomp could be
        # interpreted as the decomposition of the whole block matrix.
        
        # Check type of `covblocks` and standardize it to dictionary.
        if hasattr(covblocks, 'keys'):
            if key is not None:
                raise ValueError('can not specify key if covblocks is a dictionary')
            if None in covblocks:
                raise ValueError('None key in covblocks not allowed')
            if decomps is not None and not hasattr(decomps, 'keys'):
                raise TypeError('covblocks is dictionary but decomps is not')
        else:
            if key is None:
                raise ValueError('covblocks is not dictionary but key is None')
            covblocks = {(key, key): covblocks}
            if decomps is not None:
                decomps = {key: decomps}
        
        if decomps is None:
            decomps = {}
        
        # Convert blocks to jax arrays and determine shapes from diagonal
        # blocks.
        shapes = {}
        preblocks = {}
        for keys, block in covblocks.items():
            # TODO maybe check that keys is a 2-tuple
            for key in keys:
                if key in self._elements:
                    raise KeyError(f'key {key!r} already in GP')
            xkey, ykey = keys
            if block is None:
                raise TypeError(f'block {keys!r} is None')
                # because jnp.asarray(None) interprets None as nan
                # (see jax issue #14506)
            block = jnp.asarray(block)
            
            if xkey == ykey:
                
                if block.ndim % 2 == 1:
                    raise ValueError(f'diagonal block {key!r} has odd number of axes')
                
                half = block.ndim // 2
                head = block.shape[:half]
                tail = block.shape[half:]
                if head != tail:
                    raise ValueError(f'shape {block.shape!r} of diagonal block {key!r} is not symmetric')
                shapes[xkey] = head
                
                with _jaxext.skipifabstract():
                    if self._checksym and not jnp.allclose(block, block.T):
                        raise ValueError(f'diagonal block {key!r} is not symmetric')
                
            preblocks[keys] = block
        
        # Check decomps is consistent with covblocks.
        for key, dec in decomps.items():
            if key not in shapes:
                raise KeyError(f'key {key!r} in decomps not found in diagonal blocks')
            if not isinstance(dec, _linalg.Decomposition):
                raise TypeError(f'decomps[{key!r}] = {dec!r} is not a decomposition')
            n = math.prod(shapes[key])
            if dec.n != n:
                raise ValueError(f'decomposition matrix size {dec.n} != diagonal block size {n} for key {key!r}')
        
        # Reshape blocks to square matrices and check that the shapes of out of
        # diagonal blocks match those of diagonal ones.
        blocks = {}
        for keys, block in preblocks.items():
            with _jaxext.skipifabstract():
                if self._checkfinite and not jnp.all(jnp.isfinite(block)):
                    raise ValueError(f'block {keys!r} not finite')
            xkey, ykey = keys
            if xkey == ykey:
                size = math.prod(shapes[xkey])
                blocks[keys] = block.reshape((size, size))
            else:
                for key in keys:
                    if key not in shapes:
                        raise KeyError(f'key {key!r} from off-diagonal block {keys!r} not found in diagonal blocks')
                eshape = shapes[xkey] + shapes[ykey]
                if block.shape != eshape:
                    raise ValueError(f'shape {block.shape!r} of block {keys!r} is not {eshape!r} as expected from diagonal blocks')
                xsize = math.prod(shapes[xkey])
                ysize = math.prod(shapes[ykey])
                block = block.reshape((xsize, ysize))
                blocks[keys] = block
                revkeys = keys[::-1]
                blockT = preblocks.get(revkeys)
                if blockT is None:
                    blocks[revkeys] = block.T
        
        # Check symmetry of out of diagonal blocks.
        if self._checksym:
            with _jaxext.skipifabstract():
                for keys, block in blocks.items():
                    xkey, ykey = keys
                    if xkey != ykey:
                        blockT = blocks[ykey, xkey]
                        if not jnp.allclose(block.T, blockT):
                            raise ValueError(f'block {keys!r} is not the transpose of block {revkeys!r}')
        
        # Create _Cov objects.
        for key, shape in shapes.items():
            self._elements[key] = self._Cov(blocks, shape)
            decomp = decomps.get(key)
            if decomp is not None:
                self._decompcache[key,] = decomp

    def _makecovblock_points(self, xkey, ykey):
        x = self._elements[xkey]
        y = self._elements[ykey]
        
        assert isinstance(x, self._Points)
        assert isinstance(y, self._Points)
        
        kernel = self._crosskernel(x.proc, y.proc)
        if kernel is self._zerokernel:
            # TODO handle zero cov block efficiently
            return jnp.zeros((x.size, y.size))
        
        kernel = kernel.linop('diff', x.deriv, y.deriv)
        
        if x is y and not self._checksym and self._halfmatrix:
            ix, iy, back = self._triu_indices_and_back(x.size)
            flat = x.x.reshape(-1)
            ax = flat[ix]
            ay = flat[iy]
            halfcov = kernel(ax, ay)
            cov = halfcov[back]
            # TODO to avoid inefficiencies like in BART, maybe _Kernel should
            # have a method outer(x) that by default simply does self(x[None,
            # :], x[:, None]) but can be overwritten. This halfmatrix impl could
            # be moved there with an option outer(x, *, half=False). To carry
            # over custom implementations of outer, there should be a callable
            # attribute _outer, optionally set at initialization, that is
            # transformed by kernel operations.
        else:
            ax = x.x.reshape(-1)[:, None]
            ay = y.x.reshape(-1)[None, :]
            cov = kernel(ax, ay)
        
        return cov
    
    def _makecovblock_lintransf_any(self, xkey, ykey):
        x = self._elements[xkey]
        y = self._elements[ykey]
        assert isinstance(x, self._LinTransf)
        
        # Gather covariance matrices to be transformed.
        covs = []
        for k in x.keys:
            elem = self._elements[k]
            cov = self._covblock(k, ykey)
            assert cov.shape == (elem.size, y.size)
            cov = cov.reshape(elem.shape + (y.size,))
            covs.append(cov)
        
        # Apply transformation.
        t = jax.vmap(x.transf, -1, -1)
        cov = t(*covs)
        assert cov.shape == x.shape + (y.size,)
        return cov.reshape((x.size, y.size)) # don't leave out the ()!
        # the () probably was an obscure autograd bug, I don't think it will
        # be a problem again with jax
            
    def _makecovblock(self, xkey, ykey):
        x = self._elements[xkey]
        y = self._elements[ykey]
        if isinstance(x, self._Points) and isinstance(y, self._Points):
            cov = self._makecovblock_points(xkey, ykey)
        elif isinstance(x, self._LinTransf):
            cov = self._makecovblock_lintransf_any(xkey, ykey)
        elif isinstance(y, self._LinTransf):
            cov = self._makecovblock_lintransf_any(ykey, xkey)
            cov = cov.T
        elif isinstance(x, self._Cov) and isinstance(y, self._Cov) and x.blocks is y.blocks and (xkey, ykey) in x.blocks:
            cov = x.blocks[xkey, ykey]
        else:
            # TODO handle zero cov block efficiently
            cov = jnp.zeros((x.size, y.size))
        
        with _jaxext.skipifabstract():
            if self._checkfinite and not jnp.all(jnp.isfinite(cov)):
                raise RuntimeError(f'covariance block {(xkey, ykey)!r} is not finite')
            if self._checksym and xkey == ykey and not jnp.allclose(cov, cov.T):
                raise RuntimeError(f'covariance block {(xkey, ykey)!r} is not symmetric')

        return cov
    
    def _covblock(self, row, col):
        
        if (row, col) not in self._covblocks:
            block = self._makecovblock(row, col)
            if row != col:
                if self._checksym:
                    with _jaxext.skipifabstract():
                        blockT = self._makecovblock(col, row)
                        if not jnp.allclose(block.T, blockT):
                            msg = 'covariance block {!r} is not symmetric'
                            raise RuntimeError(msg.format((row, col)))
                self._covblocks[col, row] = block.T
            self._covblocks[row, col] = block

        return self._covblocks[row, col]
        
    def _assemblecovblocks(self, rowkeys, colkeys=None):
        if colkeys is None:
            colkeys = rowkeys
        blocks = [
            [self._covblock(row, col) for col in colkeys]
            for row in rowkeys
        ]
        return jnp.block(blocks)
    
    def _checkpos(self, cov):
        with _jaxext.skipifabstract():
            # eigv = jnp.linalg.eigvalsh(cov)
            # mineigv, maxeigv = jnp.min(eigv), jnp.max(eigv)
            with warnings.catch_warnings():
                warnings.filterwarnings('ignore', r'Exited at iteration .+? with accuracies')
                warnings.filterwarnings('ignore', r'Exited postprocessing with accuracies')
                X = numpy.random.randn(len(cov), 1)
                A = numpy.asarray(cov)
                (mineigv,), _ = sparse.linalg.lobpcg(A, X, largest=False)
                (maxeigv,), _ = sparse.linalg.lobpcg(A, X, largest=True)
            assert mineigv <= maxeigv
            if mineigv < 0:
                bound = -len(cov) * jnp.finfo(cov.dtype).eps * maxeigv * self._posepsfac
                if mineigv < bound:
                    msg = 'covariance matrix is not positive definite: '
                    msg += 'mineigv = {:.4g} < {:.4g}'.format(mineigv, bound)
                    raise numpy.linalg.LinAlgError(msg)
    
    _checkpos_cache = functools.cached_property(lambda self: [])
    def _checkpos_keys(self, keys):
        # TODO go back to ancestors of _LinTransf?
        if not self._checkpositive:
            return
        keys = set(keys)
        for prev_keys in self._checkpos_cache:
            if keys.issubset(prev_keys):
                return
        cov = self._assemblecovblocks(list(keys))
        self._checkpos(cov)
        self._checkpos_cache.append(keys)
    
    def _priorpointscov(self, key):
        
        x = self._elements[key]
        classes = (self._Points, self._Cov)
        assert isinstance(x, classes)
        mean = numpy.zeros(x.size)
        cov = self._covblock(key, key).astype(float)
        assert cov.shape == 2 * mean.shape, cov.shape
        
        # get preexisting primary gvars to be correlated with the new ones
        preitems = [
            k
            for k, px in self._elements.items()
            if isinstance(px, classes)
            and k in self._priordict
        ]
        if preitems:
            prex = numpy.concatenate([
                numpy.reshape(self._priordict[k], -1)
                for k in preitems
            ])
            precov = numpy.concatenate([
                self._covblock(k, key).astype(float)
                for k in preitems
            ])
            g = gvar.gvar(mean, cov, prex, precov, fast=True)
        else:
            g = gvar.gvar(mean, cov, fast=True)
        
        return g.reshape(x.shape)
    
    def _priorlintransf(self, key):
        x = self._elements[key]
        assert isinstance(x, self._LinTransf)
        
        # Gather all gvars to be transformed.
        elems = [
            self._prior(k).reshape(-1)
            for k in x.keys
        ]
        g = numpy.concatenate(elems)
        
        # Extract jacobian and split it.
        slices = self._slices(x.keys)
        jac, indices = _gvarext.jacobian(g)
        jacs = [
            jac[s].reshape(self._elements[k].shape + indices.shape)
            for s, k in zip(slices, x.keys)
        ]
        # TODO the jacobian can be extracted much more efficiently when the
        # elements are _Points or _Cov, since in that case the gvars are primary
        # and contiguous within each block, so each jacobian is the identity + a
        # range. Then write a function _gvarext.merge_jacobians to combine
        # them, which also can be optimized knowing the indices are
        # non-overlapping ranges.
        
        # Apply transformation.
        t = jax.vmap(x.transf, -1, -1)
        outjac = t(*jacs)
        assert outjac.shape == x.shape + indices.shape
        
        # Rebuild gvars.
        outg = _gvarext.from_jacobian(numpy.zeros(x.shape), outjac, indices)
        return outg
    
    def _prior(self, key):
        prior = self._priordict.get(key, None)
        if prior is None:
            x = self._elements[key]
            if isinstance(x, (self._Points, self._Cov)):
                prior = self._priorpointscov(key)
            elif isinstance(x, self._LinTransf):
                prior = self._priorlintransf(key)
            else: # pragma: no cover
                raise TypeError(type(x))
            self._priordict[key] = prior
        return prior
    
    def prior(self, key=None, *, raw=False):
        """
        
        Return an array or a dictionary of arrays of gvars representing the
        prior for the Gaussian process. The returned object is not unique but
        the gvars stored inside are, so all the correlations are kept between
        objects returned by different calls to `prior`.
        
        Calling without arguments returns the complete prior as a dictionary.
        If you specify ``key``, only the array for the requested key is returned.
        
        Parameters
        ----------
        key : None, key or list of keys
            Key(s) corresponding to one passed to `addx` or `addtransf`. None
            for all keys.
        raw : bool
            If True, instead of returning a collection of gvars return
            their covariance matrix as would be returned by `gvar.evalcov`.
            Default False.
        
        Returns
        -------
        If raw=False (default):
        
        prior : np.ndarray or dict
            A collection of gvars representing the prior.
        
        If raw=True:
        
        cov : np.ndarray or dict
            The covariance matrix of the prior.
        """
        raw = bool(raw)
        
        if key is None:
            outkeys = list(self._elements)
        elif isinstance(key, list):
            outkeys = key
        else:
            outkeys = None
        
        self._checkpos_keys([key] if outkeys is None else outkeys)
        
        if raw and outkeys is not None:
            return {
                (row, col):
                self._covblock(row, col).reshape(
                    self._elements[row].shape +
                    self._elements[col].shape
                )
                for row in outkeys
                for col in outkeys
            }
        elif raw:
            return self._covblock(key, key).reshape(2 * self._elements[key].shape)
        elif outkeys is not None:
            return {key: self._prior(key) for key in outkeys}
        else:
            return self._prior(key)
        
    def _slices(self, keylist):
        """
        Return list of slices for the positions of flattened arrays
        corresponding to keys in ``keylist`` into their concatenation.
        """
        sizes = [self._elements[key].size for key in keylist]
        stops = numpy.pad(numpy.cumsum(sizes), (1, 0))
        return [slice(stops[i - 1], stops[i]) for i in range(1, len(stops))]



# lsqfitgp/_jaxext/_batcher.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools
import math

from jax import lax
from jax import numpy as jnp
import numpy

def batchufunc(func, *, maxnbytes):
    """

    Make a batched version of an universal function.

    The function is modified to process its inputs in chunks.

    Parameters
    ----------
    func : callable
        A jax-traceable universal function. All positional arguments are assumed
        to be arrays which are broadcasted to determine the shape.
    maxnbytes : number
        The maximum number of bytes in each input chunck over all input arrays
        after broadcasting.

    Return
    ------
    batched_func : callable
        The batched version of `func`. Keywords arguments are passed as-is to
        the function.

    """

    maxnbytes = int(maxnbytes)
    assert maxnbytes > 0

    @functools.wraps(func)
    def batched_func(*args, **kw):

        shape = jnp.broadcast_shapes(*(arg.shape for arg in args))
        if not shape or any(size == 0 for size in shape):
            return func(*args)

        rowsize = math.prod(shape[1:])
        rownbytes = rowsize * sum(arg.dtype.itemsize for arg in args)
        totalnbytes = shape[0] * rownbytes
        if totalnbytes <= maxnbytes:
            return func(*args)

        args = [arg.reshape((1,) * (len(shape) - arg.ndim) + arg.shape) for arg in args]
        short_args_idx = [i for i, arg in enumerate(args) if arg.shape[0] == 1]
        long_args_idx = [i for i, arg in enumerate(args) if arg.shape[0] > 1]
        short_args = [args[i].squeeze(0) for i in short_args_idx]
        long_args = [args[i] for i in long_args_idx]

        def combine_args(short_args, long_args):
            args = [None] * (len(short_args) + len(long_args))
            for i, arg in zip(short_args_idx, short_args):
                args[i] = arg
            for i, arg in zip(long_args_idx, long_args):
                args[i] = arg
            return args

        if rownbytes <= maxnbytes:
            # batch over leading axis

            batchsize = maxnbytes // rownbytes
            nbatches = shape[0] // batchsize
            batchedsize = nbatches * batchsize
            
            sliced_args = [arg[:batchedsize] for arg in long_args]
            batched_args = [
                arg.reshape((nbatches, batchsize) + arg.shape[1:])
                for arg in sliced_args
            ]
            def scan_loop_body(short_args, batched_args):
                assert all(arg.ndim == len(shape) - 1 for arg in short_args)
                assert all(arg.ndim == len(shape) for arg in batched_args)
                args = combine_args(short_args, batched_args)
                out = func(*args, **kw)
                assert out.shape == (batchsize,) + shape[1:]
                return short_args, out
            _, out = lax.scan(scan_loop_body, short_args, batched_args)
            assert out.shape == (nbatches, batchsize) + shape[1:]
            out = out.reshape((batchedsize,) + shape[1:])
            
            remainder_args = [arg[batchedsize:] for arg in long_args]
            args = combine_args(short_args, remainder_args)
            remainder = func(*args)
            assert remainder.shape == (shape[0] - batchedsize,) + shape[1:]
            
            out = jnp.concatenate([out, remainder])

        else:
            # cycle over leading axis, recurse

            def scan_loop_body(short_args, long_args):
                args = combine_args(short_args, long_args)
                assert all(arg.ndim == len(shape) - 1 for arg in args)
                out = batched_func(*args, **kw)
                assert out.shape == shape[1:]
                return short_args, out
            _, out = lax.scan(scan_loop_body, short_args, long_args)

        assert out.shape == shape
        return out

    return batched_func



# lsqfitgp/copula/__init__.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" Reparametrize probability distributions as Normal """

from ._base import DistrBase
from ._distr import Distr, distribution
from ._copula import Copula
from ._makedict import makedict
from ._copulas import *

# TODO I could try to drop BufferDict altogether. It adds more complexity than
# necessary: won't keep track of dependencies between keys, structure fixed to a
# dictionary, and global unique transformation names.
#
# I need something analogous to carry around both the distribution definition
# and specific values. Copula.concretize(x) -> ConcreteCopula object. Method
# .values() returns pytree of transformed values (container-copied),
# .input_values() the flat input array (readonly view). Prints as:
# Copula({
#     'x': beta(1, 2),
#     'y': gamma(1, <x>),
# })
# ConcreteCopula({
#     'x': 0.1234,
#     'y': 0.1314,
# })
# But there is a description method to list the original values:
# ConcreteCopula({
#     'x': 0.1234 <- 0.1442,
#     'y': 0.1314 <- 0.4124, <x>
# })
# or to do comparisons:
# ConcreteCopula({
#     'x': 0.1234 | 0.12344 <- 0.1442 | 0.7777,
#     'y': 0.1314 | 0.4144 <- 0.4124 | 0.1341, <x>
# })
#
# Make empbayes_fit take Copula_like as input, and output ConcreteCopula for
# the posterior. Attribute .p would be ConcreteCopula.values(), new attribute
# post would contain everything.
#
# I could maintain the BufferDict functionality, listed under a bottom section
# in the reference.



# lsqfitgp/copula/_gamma.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

"""
JAX-compatible implementation of the gamma and related distributions
"""

import functools

from scipy import special
import jax
from jax.scipy import special as jspecial
from jax import numpy as jnp
import numpy

from .. import _jaxext

def _castto(func, type):
    @functools.wraps(func)
    def newfunc(*args, **kw):
        return func(*args, **kw).astype(type)
    return newfunc

@jax.custom_jvp
def gammainccinv(a, y):
    a = jnp.asarray(a)
    y = jnp.asarray(y)
    dtype = _jaxext.float_type(a.dtype, y.dtype)
    ufunc = _castto(special.gammainccinv, dtype)
    return _jaxext.pure_callback_ufunc(ufunc, dtype, a, y)

dQ_da = _jaxext.elementwise_grad(jspecial.gammaincc, 0)
dQ_dx = _jaxext.elementwise_grad(jspecial.gammaincc, 1)

@gammainccinv.defjvp
def gammainccinv_jvp(primals, tangents):
    a, y = primals
    at, yt = tangents
    x = gammainccinv(a, y)
    a = jnp.asarray(a)
    a = a.astype(_jaxext.float_type(a))
    # convert a to float to avoid dQ_da complaining even when we are not
    # actually deriving w.r.t. a
    dQ_da_a_x = dQ_da(a, x)
    dQ_dx_a_x = dQ_dx(a, x)
    dQinv_dy_a_y = 1 / dQ_dx_a_x
    dQinv_da_a_y = -dQinv_dy_a_y * dQ_da_a_x
    return x, dQinv_da_a_y * at + dQinv_dy_a_y * yt

@jax.custom_jvp
def gammaincinv(a, y):
    a = jnp.asarray(a)
    y = jnp.asarray(y)
    dtype = _jaxext.float_type(a.dtype, y.dtype)
    ufunc = _castto(special.gammaincinv, dtype)
    return _jaxext.pure_callback_ufunc(ufunc, dtype, a, y)

dP_da = _jaxext.elementwise_grad(jspecial.gammainc, 0)
dP_dx = _jaxext.elementwise_grad(jspecial.gammainc, 1)

@gammaincinv.defjvp
def gammaincinv_jvp(primals, tangents):
    a, y = primals
    at, yt = tangents
    x = gammaincinv(a, y)
    a = jnp.asarray(a)
    a = a.astype(_jaxext.float_type(a))
    # convert a to float to avoid dP_da complaining even when we are not
    # actually deriving w.r.t. a
    dP_da_a_x = dP_da(a, x)
    dP_dx_a_x = dP_dx(a, x)
    dPinv_dy_a_y = 1 / dP_dx_a_x
    dPinv_da_a_y = -dPinv_dy_a_y * dP_da_a_x
    return x, dPinv_da_a_y * at + dPinv_dy_a_y * yt

def _gammaisf_normcdf_large_neg_x(x, a):
    logphi = lambda x: -1/2 * jnp.log(2 * jnp.pi) - 1/2 * jnp.square(x) - jnp.log(-x)
    logq = logphi(x)
    loggammaa = jspecial.gammaln(a)
    f = lambda y: (a - 1) * jnp.log(y) - y - loggammaa - logq
    f1 = lambda y: (a - 1) / y - 1
    y0 = -logq
    y1 = y0 - ((a - 1) * jnp.log(y0) - loggammaa) / ((a - 1) / y0 - 1)
    return y1

    # TODO Improve the accuracy. I tried adding one Newton step more, but it
    # does not improve the accuracy. I probably have to add terms to the
    # approximations of Phi and Q. I could try first special.erfcx for Phi.

    # x -> -∞,  q -> 0+,  y -> ∞
    # q = Φ(x) ≈ -1/√2π exp(-x²/2)/x
    # q = Q(a, y) ≈ y^(a-1) e^-y / Γ(a)
    # gamma.isf(q, a) = Q⁻¹(a, q)
    # log q = -1/2 log 2π - x²/2 - log(-x)              (1)
    # log q = (a - 1) log y - y - log Γ(a)              (2)
    # f(y) = (a - 1) log y - y - log Γ(a) - log(q)
    #      = 0     by (2)
    # f'(y) = (a - 1) / y - 1
    # y_0 = -log q    by considering y -> ∞
    # y_1 = y_0 - f(y_0) / f'(y_0)    Newton step

def _loggammaisf_normcdf_large_neg_x(x, a):
    logphi = lambda x: -1/2 * jnp.log(2 * jnp.pi) - 1/2 * jnp.square(x) - jnp.log(-x)
    logq = logphi(x)
    loggammaa = jspecial.gammaln(a)
    g = lambda logy: (a - 1) * logy - jnp.exp(logy) - loggammaa - logq
    g1 = lambda logy: (a - 1) - jnp.exp(logy)
    logy0 = jnp.log(-logq)
    logy1 = logy0 - ((a - 1) * logy0 - loggammaa) / ((a - 1) + logq)
    return logy1

class gamma:
    
    @staticmethod
    def ppf(q, a):
        return gammaincinv(a, q)

    @staticmethod
    def isf(q, a):
        return gammainccinv(a, q)

class invgamma:
    
    @staticmethod
    def ppf(q, a):
        return 1 / gammainccinv(a, q)

    @staticmethod
    def isf(q, a):
        return 1 / gammaincinv(a, q)

    @staticmethod
    def logpdf(x, a):
        return -(a + 1) * jnp.log(x) - 1 / x - jspecial.gammaln(a)

    @staticmethod
    def cdf(x, a):
        return jspecial.gammaincc(a, 1 / x)

class loggamma:

    @staticmethod
    def ppf(q, c):
        # scipy code:
        # g = sc.gammaincinv(c, q)
        # return _lazywhere(g < _XMIN, (g, q, c),
        #                   lambda g, q, c: (np.log(q) + sc.gammaln(c+1))/c,
        #                   f2=lambda g, q, c: np.log(g))
        g = gammaincinv(c, q)
        return jnp.where(g < jnp.finfo(g.dtype).tiny,
            (jnp.log(q) + jspecial.gammaln(c + 1)) / c,
            jnp.log(g),
        )

    @staticmethod
    def isf(q, c):
        # scipy code:
        # g = sc.gammainccinv(c, q)
        # return _lazywhere(g < _XMIN, (g, q, c),
        #                   lambda g, q, c: (np.log1p(-q) + sc.gammaln(c+1))/c,
        #                   f2=lambda g, q, c: np.log(g))
        g = gammainccinv(c, q)
        return jnp.where(g < jnp.finfo(g.dtype).tiny,
            (jnp.log1p(-q) + jspecial.gammaln(c + 1)) / c,
            jnp.log(g),
        )



# lsqfitgp/_GP/__init__.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from ._gp import GP

# TODO planned rewrite:
# Methods:
#     Processes
#         .defproc(key, *args, *, deriv=0)
#             *args is either (mean, kernel) or just (kernel,). The mean is an
#             ordinary callable.
#     Inspection
#         .elements() return a read-only mapping of keys -> named tuples with
#             fields ancestors=(tuple of keys), kind=str (x, cov, transf), shape,
#             x (array or None), proc (proc key or None). The mapping generates
#             the descriptions on the fly in __getitem__.
#         .processes() analogous with ancestors, kind (kernel, transf), kernel
#             (Kernel or None)
#         .dtype the x dtype, None if no x passed yet
#         .kernel(*args) if one key a Kernel, if two a CrossKernel
#         .meanfunc(prockey)
#         .__repr__() -> multiline readable description
#     Mean functions
#         .__init__(*args, ...) supports either only the kernel or the mean and
#              the kernel of the DefaultProcess.
#         .add(*args) arrays/dicts mean and cov or just cov, replaces .addcov.
#     Conditioning
#         .condition(given, givencov=None) -> new GP, with appended new keys and
#             values to condition on. given and givencov keep dict layout like
#             now.
#         .extend(given, givencov=None) -> new GP, like current predfromfit. It
#             sets a flag about the kind of conditioning in the list of given
#             stuff. I have to derive how to condition and extend together.
#     Distribution access
#         .mean(key) -> array, the key can be a pytree
#         .cov(*args) -> if one arg: square covariance, if two args: cross
#             covariance. The keys can be pytrees, and the result is the pytree
#             product with matrices as leaves.
#         .gvars(key, keepcorr=True) -> tree of arrays of gvars. keepcorr=True
#             won't work if condition was not passed gvars, so this default raises
#             an error if that happens.
#         .sample(key, n=1) -> array, key can be a pytree. caches the decomp?
#     Density
#         ._decomp(y, ycov=None) -> resid, cov (like _prior_decomp now)
#         .logpdf(y, ycov=None) -> like marginal_likelihood now

# TODO methods to implement linear models. The elegant way to do this is with
# kernels taking a formula, but that would not take advantage of the fact that
# the kernel is low-rank if I re-implement Woodbury. I need to define an element
# for the coefficients, and then the process works by definining finite
# transformations of this element instead of going through the usual kernel
# route.
#
# A more general route would be `features` mechanism for kernels: a kernel can
# provide itself a way to evaluate a covariance matrix as a low-rank product.



# lsqfitgp/_special/_zeta.py
#
# Copyright (c) 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import collections
import functools
import math

import jax
from jax import lax
from jax import numpy as jnp
from jax.scipy import special as jspecial

from .. import _jaxext
from . import _gamma

def hurwitz_zeta_series(m, x, a1, onlyeven=False, onlyodd=False, skipterm=None):
    """
    hurwitz zeta(s = m + x, a = 1 - a1) with integer m
    meant to be used with |x| ≤ 1/2, but no actual restriction
    assuming -S <= s <= 0 and |a1| <= 1/2 with S ~ some decade
    https://dlmf.nist.gov/25.11.E10
    """
    
    # decide number of terms to sum
    t = _jaxext.float_type(m, x, a1)
    nmax = hze_nmax(t)
    n = jnp.arange(nmax + 1)
    
    # make arguments broadcastable with n
    x = x[..., None]
    m = m[..., None]
    a1 = a1[..., None]
    if skipterm is not None:
        skipterm = skipterm[..., None]
    
    # compute pochhammer symbol, factorial and power terms
    nm = n + m
    ns1 = nm - 1 + x # = n + s - 1
    ns1_limit = jnp.where(ns1 == 0, 1, ns1) # pochhammer zero cancels zeta pole
    ns1_limit = jnp.where(ns1 == 1, 0, ns1_limit)
    # TODO this == 1 worries me, maybe sometimes it's violated, use shift
    factor = jnp.cumprod((ns1_limit * a1 / n).at[..., 0].set(1), -1, t)
    
    # handle tweaks to the series
    if onlyeven:
        sl = slice(None, None, 2)
    elif onlyodd:
        sl = slice(1, None, 2)
    if onlyeven or onlyodd:
        n = n[sl]
        nm = nm[..., sl]
        ns1 = ns1[..., sl]
        factor = factor[..., sl]
    if skipterm is not None:
        factor = jnp.where(n == skipterm, 0, factor)
    
    # compute zeta term
    zet = zeta(x, nm) # = zeta(n + s)
    zet_limit = jnp.where(ns1 == 0, 1, zet) # pole cancelled by pochhammer
    
    # sum series
    kw = dict(precision=lax.Precision.HIGHEST)
    series = jnp.matmul(factor[..., None, :], zet_limit[..., :, None], **kw)
    return series.squeeze((-2, -1))

def hze_nmax(t):
    minz = 0.0037 # = min(2 gamma(s) / (2 pi)^s) for s <= 0
    return int(math.ceil(-math.log2(jnp.finfo(t).eps * minz)))

# @jax.jit
def hurwitz_zeta(s, a):
    """
    For 0 <= a <= 1 and -S <= s <= 0 with S not too large
    """
    s = jnp.asarray(s)
    a = jnp.asarray(a)
    
    cond = a < 1/2  # do a + 1 to bring a closer to 1
    a1 = jnp.where(cond, -a, 1. - a)
    zero = jnp.array(0)
    zeta = hurwitz_zeta_series(zero, s, a1)  # https://dlmf.nist.gov/25.11.E10
    zeta += jnp.where(cond, a ** -s, 0)  # https://dlmf.nist.gov/25.11.E3
    return zeta
    
    # https://specialfunctions.juliamath.org/stable/functions_list/#SpecialFunctions.zeta

@functools.partial(jax.custom_jvp, nondiff_argnums=(1, 2))
# @functools.partial(jax.jit, static_argnums=(2,))
def periodic_zeta(x, s, imag=False):
    """
    compute F(x,s) = Li_s(e^2πix) for real s > 1, real x
    """
    
    x = jnp.asarray(x)
    s = jnp.asarray(s)
    
    # decide boundary for large/small s implementation
    t = _jaxext.float_type(x, s)
    eps = jnp.finfo(t).eps
    nmax = 50
    larges = math.ceil(-math.log(eps) / math.log(nmax)) # 1/nmax^s < eps
    
    z_smalls = periodic_zeta_smalls(x, s, imag)
    z_larges = periodic_zeta_larges(x, s, nmax, imag)

    return jnp.where(s < larges, z_smalls, z_larges)
    
@periodic_zeta.defjvp
def periodic_zeta_jvp(s, imag, p, t):
    x, = p
    xt, = t
    primal = periodic_zeta(x, s, imag)
    sgn = 1 if imag else -1
    tangent = 2 * jnp.pi * sgn * periodic_zeta(x, s - 1, not imag) * xt
    return primal, tangent

def standard_x(x):
    """ bring x in [0, 1/2] by modulus and reflection """
    x %= 1
    neg = x > 1/2
    return neg, jnp.where(neg, 1 - x, x)

def periodic_zeta_larges(x, s, nmax, imag):
    """ https://dlmf.nist.gov/25.13.E1 """

    t = _jaxext.float_type(x, s)
    s = s.astype(t) # avoid n^s overflow with integer s
    n = jnp.arange(1, nmax + 1)
    neg, nx = standard_x(n * x[..., None])
    func = jnp.sin if imag else jnp.cos
    terms = func(2 * jnp.pi * nx) / n ** s[..., None]
    if imag:
        terms *= jnp.where(neg, -1, 1)
    return jnp.sum(terms, -1)

def periodic_zeta_smalls(x, s, imag):
    """
    https://dlmf.nist.gov/25.11.E10 and https://dlmf.nist.gov/25.11.E3 expanded
    into https://dlmf.nist.gov/25.13.E2
    """
    neg, x = standard_x(x) # x in [0, 1/2]
    
    eps = jnp.finfo(_jaxext.float_type(x, s)).eps
    s = jnp.where(s % 1, s, s * (1 + eps)) # avoid integer s

    s1 = 1 - s  # < 0
    q = -jnp.around(s1).astype(int)
    a = s1 + q
    # now s1 == -q + a with q integer >= 0 and |a| <= 1/2
    
    pi = (2 * jnp.pi) ** -s1
    gam = _gamma.gamma(s1)
    func = sin_pi2 if imag else cos_pi2
    pha = func(-q, a) # = sin or cos(π/2 s1), numerically accurate for small a
    hzs = 2 * hurwitz_zeta_series(-q, a, -x, onlyeven=not imag, onlyodd=imag, skipterm=q)
    # hzs = ζ(s1,1+x) -/+ ζ(s1,1-x) but without the x^q term in the series
    pdiff = zeta_series_power_diff(x, q, a)
    # pdiff accurately handles the sum of the external power x^-s1 due to
    # 25.11.E3 with the q-th term (cancellation) with even q for the real part
    # and odd q for the imaginary part
    cancelcond = jnp.logical_and(imag, q % 2 == 1)
    cancelcond |= jnp.logical_and(not imag, q % 2 == 0)
    power = jnp.where(cancelcond, pdiff, x ** -s1)
    hz = power + hzs # = ζ(s1,x) -/+ ζ(s1,1-x)
    
    out = (pi * gam * pha) * hz
    if imag:
        out *= jnp.where(neg, -1, 1)
    return out

def cos_pi2(n, x):
    """ compute cos(π/2 (n + x)) for n integer, accurate for small x """
    arg = -jnp.pi / 2 * x
    cos = jnp.where(n % 2, jnp.sin(arg), jnp.cos(arg))
    return cos * jnp.where(n // 2 % 2, -1, 1)

def sin_pi2(n, x):
    return cos_pi2(n - 1, x)

def zeta_series_power_diff(x, q, a):
    """
    compute x^q-a + (-1)^q * [q-th term of 2 * hurwitz_zeta_series(-q, a, x)]
    """
    pint = x ** q
    pz = jnp.where(q, 0, jnp.where(a, -1, 0)) # * 0^q = 0^q-a - 0^q
    pdif = jnp.where(x, jnp.expm1(-a * jnp.log(x)), pz) # * x^q = x^q-a - x^q
    gamincr = jnp.where(q, _gamma.gamma_incr(1 + q, -a), 0)
    # gamincr = Γ(1+q-a) / Γ(1+q)Γ(1-a)  -  1
    zz = zeta_zero(a) # = ζ(a) - ζ(0)
    qdif = 2 * (1 + gamincr) * zz - gamincr # = (q-th term) - (q-th term)|_a=0
    return pint * (pdif + qdif)

def zeta_zero(s):
    """
    Compute zeta(s) - zeta(0) for |s| < 1 accurately
    """
    
    # f(s) = zeta(s) - 1 / (s - 1)
    # I have the Taylor series of f(s)
    # zeta(s) - zeta(0) = f(s) + 1 / (s - 1) + 1/2 =
    # = f(s) + 1/(s-1) + 1 - 1 + 1/2 =
    # = f(s) - 1/2 + s/(s-1)
    
    t = _jaxext.float_type(s)
    coef = jnp.array(zeta_zero_coef, t).at[0].set(0)
    fact = jnp.cumprod(jnp.arange(coef.size).at[0].set(1), dtype=t)
    coef /= fact
    f = jnp.polyval(coef[::-1], s)
    return f + s / (s - 1)

zeta_zero_coef = [         # = gen_zeta_zero_coef(17)
    0.5,
    0.08106146679532726,
    -0.006356455908584851,
    -0.004711166862254448,
    0.002896811986292041,
    -0.00023290755845472455,
    -0.0009368251300509295,
    0.0008498237650016692,
    -0.00023243173551155957,
    -0.00033058966361229646,
    0.0005432341157797085,
    -0.00037549317290726367,
    -1.960353628101392e-05,
    0.00040724123256303315,
    -0.0005704920132817777,
    0.0003939270789812044,
    8.345880582550168e-05,
]

def gen_zeta_zero_coef(n): # pragma: no cover
    """
    Compute first n derivatives of zeta(s) - 1/(s-1) at s = 0
    """
    import mpmath as mp
    with mp.workdps(32):
        func = lambda s: mp.zeta(s) - 1 / (s - 1)
        return [float(mp.diff(func, 0, k)) for k in range(n)]

# @jax.jit
def zeta(s, n=0):
    """ compute ζ(n + s) with integer n, accurate for even n < 0 and small s """
    s = jnp.asarray(s)
    return jnp.where(n + s >= 0,
        zeta_0_inf(n + s),
        zeta_neg(s, n),
    )

def zeta_neg(s, n):
    # reflection formula https://dlmf.nist.gov/25.4.E1
    m = 1 - n
    x = -s
    # m + x = 1 - (n + s) = 1 - n - s
    mx = m + x # > 1
    logpi = -mx * jnp.log(2 * jnp.pi)
    cos = cos_pi2(m, x) # = cos(π/2 (m + x)) but accurate for small x
    loggam = jspecial.gammaln(mx)
    zet = zeta_0_inf(mx)
    
    # cancel zeta pole at 1
    cos = jnp.where(mx == 1, -jnp.pi / 2, cos)
    zet = jnp.where(mx == 1,           1, zet)
        
    return 2 * jnp.exp(logpi + loggam) * cos * zet

##########################################################################
#    The following is adapted from gsl/specfunc/zeta.c (GPL license)     #
##########################################################################

ChebSeries = collections.namedtuple('ChebSeries', 'c a b')

def cheb_eval_e(cs, x):
    d = 0.0
    dd = 0.0
    y = (2.0 * x - cs.a - cs.b) / (cs.b - cs.a)
    y2 = 2.0 * y
    
    for c in cs.c[:0:-1]:
        d, dd = y2 * d - dd + c, d

    return y * d - dd + 0.5 * cs.c[0]

# chebyshev fit for (s(t)-1)Zeta[s(t)]
# s(t)= (t+1)/2
# -1 <= t <= 1
zeta_xlt1_cs = ChebSeries(jnp.array([
     1.48018677156931561235192914649,
     0.25012062539889426471999938167,
     0.00991137502135360774243761467,
    -0.00012084759656676410329833091,
    -4.7585866367662556504652535281e-06,
     2.2229946694466391855561441361e-07,
    -2.2237496498030257121309056582e-09,
    -1.0173226513229028319420799028e-10,
     4.3756643450424558284466248449e-12,
    -6.2229632593100551465504090814e-14,
    -6.6116201003272207115277520305e-16,
     4.9477279533373912324518463830e-17,
    -1.0429819093456189719660003522e-18,
     6.9925216166580021051464412040e-21,
]), -1, 1)

# chebyshev fit for (s(t)-1)Zeta[s(t)]
# s(t)= (19t+21)/2
# -1 <= t <= 1
zeta_xgt1_cs = ChebSeries(jnp.array([
    19.3918515726724119415911269006,
     9.1525329692510756181581271500,
     0.2427897658867379985365270155,
    -0.1339000688262027338316641329,
     0.0577827064065028595578410202,
    -0.0187625983754002298566409700,
     0.0039403014258320354840823803,
    -0.0000581508273158127963598882,
    -0.0003756148907214820704594549,
     0.0001892530548109214349092999,
    -0.0000549032199695513496115090,
     8.7086484008939038610413331863e-6,
     6.4609477924811889068410083425e-7,
    -9.6749773915059089205835337136e-7,
     3.6585400766767257736982342461e-7,
    -8.4592516427275164351876072573e-8,
     9.9956786144497936572288988883e-9,
     1.4260036420951118112457144842e-9,
    -1.1761968823382879195380320948e-9,
     3.7114575899785204664648987295e-10,
    -7.4756855194210961661210215325e-11,
     7.8536934209183700456512982968e-12,
     9.9827182259685539619810406271e-13,
    -7.5276687030192221587850302453e-13,
     2.1955026393964279988917878654e-13,
    -4.1934859852834647427576319246e-14,
     4.6341149635933550715779074274e-15,
     2.3742488509048340106830309402e-16,
    -2.7276516388124786119323824391e-16,
     7.8473570134636044722154797225e-17
]), -1, 1)

def zeta_0_1(s):
    return cheb_eval_e(zeta_xlt1_cs, 2.0 * s - 1.0) / (s - 1.0)

def zeta_1_20(s):
    return cheb_eval_e(zeta_xgt1_cs, (2.0 * s - 21.0) / 19.0) / (s - 1.0)

def zeta_20_inf(s):
    f2 = 1.0 - 2.0 ** -s
    f3 = 1.0 - 3.0 ** -s
    f5 = 1.0 - 5.0 ** -s
    f7 = 1.0 - 7.0 ** -s
    return 1.0 / (f2 * f3 * f5 * f7);

def zeta_0_inf(s):
    return jnp.where(s >= 20,
        zeta_20_inf(s),
        jnp.where(s >= 1,
            zeta_1_20(s),
            zeta_0_1(s),
        ),
    )

##########################################################################



# lsqfitgp/_signature.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

""" define Signature """

import inspect

from numpy.lib import function_base
import jax
from jax import numpy as jnp

class Signature:
    """ Class to parse a numpy gufunc signature. """

    def __init__(self, signature):
        self.signature = signature
        self.incores, self.outcores = function_base._parse_gufunc_signature(signature)

    @classmethod
    def from_tuples(cls, incores, outcores):
        self = cls.__new__(cls)
        tuplestr = lambda t: '(' + ','.join(map(str, t)) + ')'
        self.signature = ','.join(map(tuplestr, incores)) + '->' + ','.join(map(tuplestr, outcores))
        self.incores = incores
        self.outcores = outcores
        return self

    def __repr__(self):
        return self.signature

    def check_nargs(self, func):
        """ Check that the function has the correct number of arguments. """
        sig = inspect.signature(func)
        if any(p.kind == inspect.Parameter.VAR_POSITIONAL for p in sig.parameters.values()):
            return
        if len(sig.parameters) != len(self.incores):
            raise ValueError(f'function {func} has {len(sig.parameters)} '
                f'arguments, but signature {self.signature} '
                f'requires {len(self.incores)}')

    @property
    def nin(self):
        return len(self.incores)

    @property
    def nout(self):
        return len(self.outcores)

    def eval(self, *args):
        """

        Evaluate the signature with the given arguments.

        Parameters
        ----------
        args : sequence of numpy.ndarray or None
            A missing argument can be replaced with None, provided the other
            arguments are sufficient to infer all dimension sizes.

        Returns
        -------
        sig : EvaluatedSignature
            An object with attributes `broadcast_shape`, `sizes`,
            `core_out_shapes`, `out_shapes`, `core_in_shapes`, `in_shapes`.

        """
        return self.EvaluatedSignature(self, *args)
    
    class EvaluatedSignature:

        def __init__(self, sig, *args):

            assert len(args) == len(sig.incores)
            
            known_args = []
            known_cores = []
            missing_cores = []
            for arg, core in zip(args, sig.incores):
                if arg is None:
                    missing_cores.append(core)
                else:
                    known_args.append(jax.ShapeDtypeStruct(arg.shape, 'd'))
                    known_cores.append(core)
            
            self.broadcast_shape, self.sizes = function_base._parse_input_dimensions(known_args, known_cores)

            missing_indices = set(sum(missing_cores, ()))
            missing_indices.difference_update(self.sizes)
            if missing_indices:
                raise ValueError(f'cannot infer sizes of dimesions {missing_indices} from signature {sig.signature}')

            self.core_out_shapes, self.out_shapes = self._compute_shapes(sig.outcores)
            self.core_in_shapes, self.in_shapes = self._compute_shapes(sig.incores)

        def _compute_shapes(self, cores):
            coreshapes = []
            shapes = []
            for core in cores:
                core = tuple(self.sizes[i] for i in core)
                coreshapes.append(core)
                shapes.append(self.broadcast_shape + core)
            return tuple(coreshapes), tuple(shapes)

        def _repr(self, shapes):
            return ','.join(map(str, shapes)).replace(' ', '')

        def __repr__(self):
            return self._repr(self.in_shapes) + '->' + self._repr(self.out_shapes)

# I use numpy's internals to parse the signature, but these do not correspond to
# the description in
# https://numpy.org/doc/stable/reference/c-api/generalized-ufuncs.html. In
# particular:
# - integers are parsed like identifiers
# - ? is not accepted
# Looking at the issues, it seems a long standing issue amongst many with
# vectorize and is not going to be solved.
# See https://github.com/HypothesisWorks/hypothesis/blob/4e675dee1a4cba9d6902290bbc5719fd72072ec7/hypothesis-python/src/hypothesis/extra/_array_helpers.py#L289
# for a more complete implementation



# lsqfitgp/_Kernel/_kernel.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

from . import _util
from . import _crosskernel

class Kernel(_crosskernel.CrossKernel):
    r"""

    Subclass of `CrossKernel` to represent the kernel of a single function:

    .. math::
        \mathrm{kernel}(x, y) = \mathrm{Cov}[f(x), f(y)].

    """

    def _swap(self):
        return self

Kernel.inherit_transf('xtransf')
Kernel.inherit_transf('diff')
# other transformations are registered by IsotropicKernel

@Kernel.register_transf
def forcekron(tcls, self):
    r"""
    
    Force the kernel to be a separate product over dimensions:

    .. math::
        \mathrm{newkernel}(x, y) = \prod_i \mathrm{kernel}(x_i, y_i)
    
    Returns
    -------
    newkernel : Kernel
        The transformed kernel.

    """

    core = self.core
    newcore = lambda x, y, **kw: _util.prod_recurse_dtype(core, x, y, **kw)
    return self._clone(tcls, core=newcore)

_crosskernel.Kernel = Kernel



# lsqfitgp/_GP/_base.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools

import jax
from jax import numpy as jnp

from .. import _jaxext
from .. import _utils

class GPBase:

    def __init__(self, *, checkfinite=True, checklin=True):
        self._checkfinite = bool(checkfinite)
        self._checklin = bool(checklin)

    def _clone(self):
        newself = object.__new__(self.__class__)
        newself._checkfinite = self._checkfinite
        newself._checklin = self._checklin
        return newself

    class _SingletonMeta(type):
    
        def __repr__(cls):
            return cls.__name__

    class _Singleton(metaclass=_SingletonMeta):
    
        def __new__(cls):
            raise NotImplementedError(f"{cls.__name__} can not be instantiated")

    class DefaultProcess(_Singleton):
        """ Key of the default process. """
        pass

    def _checklinear(self, func, inshapes, elementwise=False):
        
        # Make input arrays.
        rkey = jax.random.PRNGKey(202206091600)
        inp = []
        for shape in inshapes:
            rkey, subkey = jax.random.split(rkey)
            inp.append(jax.random.normal(subkey, shape))
        
        # Put zeros into the arrays to check they are preserved.
        if elementwise:
            shape = jnp.broadcast_shapes(*inshapes)
            rkey, subkey = jax.random.split(rkey)
            zeros = jax.random.bernoulli(subkey, 0.5, shape)
            for i, a in enumerate(inp):
                inp[i] = a.at[zeros].set(0)
            
        # Compute JVP and check it is identical to the function itself.
        with _jaxext.skipifabstract():
            out0, out1 = jax.jvp(func, inp, inp)
            if out1.dtype == jax.float0:
                cond = jnp.allclose(out0, 0)
            else:
                cond = jnp.allclose(out0, out1)
            if not cond:
                raise RuntimeError('the transformation is not linear')
        
            # Check that the function is elementwise.
            if elementwise:
                if out0.shape != shape or not (jnp.allclose(out0[zeros], 0) and jnp.allclose(out1[zeros], 0)):
                    raise RuntimeError('the transformation is not elementwise')
    
def newself(meth):
    """ Decorator to create a new GP object and pass it to the method. """

    @functools.wraps(meth)
    def newmeth(self, *args, **kw):
        self = self._clone()
        meth(self, *args, **kw)
        return self

    # append return value description to docstring
    doctail = """\
    Returns
    -------
    gp : GP
        A new GP object with the applied modifications.
    """
    newmeth.__doc__ = _utils.append_to_docstring(meth.__doc__, doctail)

    return newmeth



# lsqfitgp/_Kernel/_isotropic.py
#
# Copyright (c) 2020, 2022, 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import sys

from jax import numpy as jnp

from .. import _jaxext

from . import _util
from . import _crosskernel
from . import _kernel
from . import _stationary

class CrossIsotropicKernel(_stationary.CrossStationaryKernel):
    """
    
    Subclass of `CrossStationaryKernel` for isotropic kernels.

    An isotropic kernel depends only on the Euclidean distance between its two
    arguments.

    Parameters
    ----------
    core : callable
        A function taking one argument ``r2`` which is the squared distance
        between x and y, plus optionally keyword arguments.
    input : {'squared', 'abs', 'posabs', 'raw'}, default 'squared'
        If ``'squared'``, `core` is passed the squared distance. If ``'abs'``,
        it is passed the distance (not squared). If ``'posabs'``, it is passed
        the distance, and the distance of equal points is a small number instead
        of zero. If ``'raw'``, `core` is passed both points separately like
        non-stationary kernels.
    **kw
        Additional keyword arguments are passed to the `CrossKernel`
        constructor.
    
    Notes
    -----
    The ``input='posabs'`` option will cause problems with second derivatives in
    more than one dimension.
            
    """
     
    def __new__(cls, core, *, input='squared', **kw):
        if input == 'raw':
            return _crosskernel.CrossKernel.__new__(cls, core, **kw)
        
        if input in ('squared', 'abs'):
            dist = lambda x, y: jnp.square(x - y)
        elif input == 'posabs':
            dist = lambda x, y: jnp.square(_stationary._softabs(x - y))
        else:
            raise KeyError(input)
                
        if input in ('posabs', 'abs'):
            transf = jnp.sqrt
        else:
            transf = lambda ss: ss
        
        def newcore(x, y, **kwargs):
            ss = _util.sum_recurse_dtype(dist, x, y)
            return core(transf(ss), **kwargs)
        
        return _crosskernel.CrossKernel.__new__(cls, newcore, **kw)

    # TODO add a `distance` parameter to pick arbitrary distances, but since the
    # distance definition can not be changed arbitrarily, it may be better to
    # keep this class for the 2-norm and eventually add another if needed.
    
    # TODO it is not efficient that the distance is computed separately for
    # each kernel in a kernel expression, but probably it would be difficult
    # to support everything without bugs while also computing the distance once.
    # A possible way is adding a keyword argument to the _kernel member
    # that kernels use to memoize things, the first IsotropicKernel that gets
    # called puts the distance there. Possible name: _cache.
    
class IsotropicKernel(CrossIsotropicKernel, _stationary.StationaryKernel):
    pass

IsotropicKernel.inherit_all_algops(intermediates=True)
IsotropicKernel.inherit_transf('rescale', intermediates=True)
IsotropicKernel.inherit_transf('loc', intermediates=True)
IsotropicKernel.inherit_transf('scale', intermediates=True)
IsotropicKernel.inherit_transf('maxdim', intermediates=True)
IsotropicKernel.inherit_transf('derivable', intermediates=True)
IsotropicKernel.inherit_transf('normalize', intermediates=True)
IsotropicKernel.inherit_transf('cond', intermediates=True)

class CrossConstant(CrossIsotropicKernel):
    pass

class Constant(CrossConstant, IsotropicKernel):
    pass

def zero(x, y):
    return jnp.broadcast_to(0., jnp.broadcast_shapes(x.shape, y.shape))

class Zero(IsotropicKernel):
    """
    Represents a kernel that unconditionally yields zero.
    """

    def __new__(cls):
        return super().__new__(cls, zero, input='raw')

_crosskernel.IsotropicKernel = IsotropicKernel
_crosskernel.CrossIsotropicKernel = CrossIsotropicKernel
_crosskernel.Constant = Constant
_crosskernel.CrossConstant = CrossConstant



# lsqfitgp/bayestree/_bcf.py
#
# Copyright (c) 2023, Giacomo Petrillo
#
# This file is part of lsqfitgp.
#
# lsqfitgp is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lsqfitgp is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lsqfitgp.  If not, see <http://www.gnu.org/licenses/>.

import functools

import numpy
from scipy import stats
from jax import numpy as jnp
import jax
import gvar

from .. import copula
from .. import _kernels
from .. import _fit
from .. import _array
from .. import _GP
from .. import _fastraniter

# TODO add methods or options to do causal inference stuff, e.g., impute missing
# outcomes, or ate, att, cate, catt, sate, satt. Remember that the effect may
# also depend on aux. See bartCause, possibly copy its naming.

def _recursive_cast(dtype, default, mapping):
    if dtype in mapping:
        return mapping[dtype]
    elif dtype.names is not None:
        return numpy.dtype([
            (name, _recursive_cast(dtype[name], default, mapping))
            for name in dtype.names
        ])
    elif dtype.subdtype is not None:
        # note: has names => does not have subdtype
        return numpy.dtype((_recursive_cast(dtype.base, default, mapping), dtype.shape))
    elif default is None:
        return dtype
    else:
        return default

def cast(dtype, default, mapping={}):
    """
    Recursively cast a numpy data type.

    Parameters
    ----------
    dtype : dtype
        The data type to cast.
    default : dtype or None
        The leaf fields of `dtype` are casted to `default`, which can be
        structured, unless they appear in `mapping`. If None, dtypes not in
        `mapping` are left unchanged.
    mapping : dict
        A dictionary from dtypes to dtypes, indicating specific casting rules.
        The dtypes can be structured, a match of a structured dtype takes
        precedence over matches in its leaves, and the converted dtype is not
        further searched for matches.

    Returns
    -------
    casted_dtype : dtype
        The casted version of `dtype`. May not have the same structure if
        `mapping` contains structured dtypes.
    """
    mapping = {numpy.dtype(k): numpy.dtype(v) for k, v in mapping.items()}
    default = None if default is None else numpy.dtype(default)
    return _recursive_cast(numpy.dtype(dtype), default, mapping)

class bcf:
    
    def __init__(self, *,
        y,
        z,
        x_mu,
        x_tau=None,
        pihat,
        include_pi='mu',
        weights=None,
        fitkw={},
        kernelkw_mu={},
        kernelkw_tau={},
        marginalize_mean=True,
        gpaux=None,
        x_aux=None,
        otherhp={},
    ):
        r"""
        Nonparametric Bayesian regression with a GP version of BCF.

        BCF (Bayesian Causal Forests) is a regression method for observational
        causal inference studies introduced in [1]_ based on a pair of BART
        models.

        This class evaluates a Gaussian process regression with a kernel which
        accurately approximates BCF in the infinite trees limit of each BART
        model. The hyperparameters are optimized to their marginal MAP.

        The model is (loosely, see notes below) :math:`y = \mu(x) + z\tau(x)`,
        so :math:`\tau(x)` is the expected causal effect of :math:`z` on
        :math:`y` at location :math:`x`.

        Parameters
        ----------
        y : (n,) array, series or dataframe
            Outcome.
        z : (n,) array, series or dataframe
            Binary treatment status: 0 control group, 1 treatment group.
        x_mu : (n, p) array, series or dataframe
            Covariates for the :math:`\mu` model.
        x_tau : (n, q) array, series or dataframe, optional
            Covariates for the :math:`\tau` model. If not specified, use `x_mu`.
        pihat : (n,) array, series or dataframe
            Estimated propensity score, i.e., P(Z=1|X).
        include_pi : {'mu', 'tau', 'both'}, optional
            Whether to include the propensity score in the :math:`\mu` model,
            the :math:`\tau` model, or both. Default is ``'mu'``.
        weights : (n,) array, series or dataframe
            Weights used to rescale the error variance (as 1 / weight).
        fitkw : dict
            Additional arguments passed to `~lsqfitgp.empbayes_fit`, overrides
            the defaults.
        kernelkw_mu, kernelkw_tau : dict
            Additional arguments passed to `~lsqfitgp.BART` for each model,
            overrides the defaults.
        marginalize_mean : bool
            If True (default), marginalize the intercept of the model.
        gpaux : callable, optional
            If specified, this function is called with a pair ``(hp, gp)``,
            where ``hp`` is a dictionary of hyperparameters, and ``gp`` is a
            `~lsqfitgp.GP` object under construction, and is expected to return
            a modified ``gp`` with a new process named ``'aux'`` defined with
            `~lsqfitgp.GP.defproc` or similar. The process is added to the
            regression model. The input to the process is a structured array
            with fields:

            'train' : bool
                Indicates wether the data is training set (the one passed on
                initialization) or test set (the one passed to `pred` or `gp`).
            'i' : int
                Index of the flattened array.
            'z' : int
                Treatment status.
            'mu', 'tau' : structured
                The values in `x_mu` and `x_tau`, converted to indices according
                to the BART grids. Where `pihat` has been added, there are two
                subfields: ``'x'`` which contains the covariates, and
                ``'pihat'``, the latter expressed in indices as well.
            'pihat' : float
                The `pihat` argument. Contrary to the subfield included under
                ``'mu'`` and/or ``'tau'``, this field contains the original
                values.
            'aux' : structured
                The values in `x_aux`, if specified.
        
        x_aux : (n, k) array, series or dataframe, optional
            Additional covariates for the ``'aux'`` process.
        otherhp : dictionary of gvar
            A dictionary with the prior of additional hyperpameters, intended to
            be used by ``gpaux``.
        
        Notes
        -----
        The regression model is:

        .. math::
            y_i &= m + {} \\
                &\phantom{{}={}} +
                    \lambda_\mu\,\mathrm{std}(\mathbf y)
                    \mu(\mathbf x^\mu_i, \hat\pi_i?) + {} \\
                &\phantom{{}={}} +
                    \lambda_\tau\,\mathrm{std}(\mathbf y)
                    \tau(\mathbf x^\tau_i, \hat\pi_i?) (z_i - z_0) + {} \\
                &\phantom{{}={}} +
                    \mathrm{aux}(i, z_i, \mathbf x^\mu_i, \mathbf x^\tau_i,
                        \hat\pi_i, \mathbf x^\text{aux}_i) + {} \\
                &\phantom{{}={}} +
                    \varepsilon_i, \\
            \varepsilon_i &\sim
                N(0, \sigma^2 / w_i), \\
            m &\sim N(
                (\max(\mathbf y) + \min(\mathbf y)) / 2,
                (\max(\mathbf y) - \min(\mathbf y))^2 / 4
            ), \\
            \log \sigma^2 &\sim N(
                \log(\overline{w(y - \bar y)^2}),
                4
            ), \\
            \lambda_\mu
                &\sim \mathrm{HalfCauchy}(2), \\
            \lambda_\tau
                &\sim \mathrm{HalfNormal}(1.48), \\
            \mu &\sim \mathrm{GP}(0,
                \mathrm{BART}(\alpha_\mu, \beta_\mu) ), \\
            \tau &\sim \mathrm{GP}(0,
                \mathrm{BART}(\alpha_\tau, \beta_\tau) ), \\
            \mathrm{aux} & \sim \mathrm{GP}(0, \text{<user defined>}), \\
            \alpha_\mu, \alpha_\tau &\sim \mathrm{B}(2, 1), \\
            \beta_\mu, \beta_\tau &\sim \mathrm{IG}(1, 1), \\
            z_0 &\sim U(0, 1),

        To make the inference, :math:`(\mu, \tau, \boldsymbol\varepsilon, m,
        \mathrm{aux})` are marginalized analytically, and the marginal posterior
        mode of
        :math:`(\sigma, \lambda_*, \alpha_*, \beta_*, z_0)` is found by
        numerical minimization, after transforming them to express their prior
        as a Gaussian copula. Their marginal posterior covariance matrix is
        estimated with an approximation of the hessian inverse. See
        `~lsqfitgp.empbayes_fit` and use the parameter ``fitkw`` to customize
        this procedure.

        The tree splitting grid of the BART kernel is set using quantiles of the
        observed covariates. This corresponds to settings ``usequants=True``,
        ``numcut=inf`` in the R packages BayesTree and BART. Use the parameters
        `kernelkw_mu` and `kernelkw_tau` to customize the grids.

        The difference between the regression model evaluated at :math:`Z=1` vs.
        :math:`Z=0` can be interpreted as the causal effect :math:`Z \rightarrow
        Y` if the unconfoundedness assumption is made:

        .. math::
            \{Y(Z=0), Y(Z=1)\} \perp\!\!\!\perp Z \mid X.

        In practical terms, this holds when:

            1) :math:`X` are pre-treatment variables, i.e., they represent
               quantities causally upstream of :math:`Z`.

            2) :math:`X` are sufficient to adjust for all common causes of
               :math:`Z` and :math:`Y`, such that the only remaining difference
               is the causal effect and not just a correlation.

        Here :math:`X` consists in `x_tau`, `x_mu` and `x_aux`. However these
        arrays may also be used to pass "technical" values used to set up the
        model, that do not satisfy the uncounfoundedness assumption, if you know
        what you are doing.

        Attributes
        ----------
        m : float or gvar
            The prior mean :math:`m`.
        sigma : gvar
            The error term standard deviation :math:`\sigma`. If there are
            weights, the sdev for each unit is obtained dividing ``sigma`` by
            sqrt(weight).
        alpha_mu, alpha_tau : gvar
            The numerator of the tree spawn probability :math:`\alpha_*` (named
            ``base`` in R bcf).
        beta_mu, beta_tau : gvar
            The depth exponent of the tree spawn probability :math:`\beta_*`
            (named ``power`` in R bcf).
        sdev_mu, sdev_tau : gvar
            The prior standard deviation :math:`\lambda_* \mathrm{std}(\mathbf
            y)`.
        z_0 : gvar
            The treatment coding parameter.
        fit : empbayes_fit
            The hyperparameters fit object.

        Methods
        -------
        gp :
            Create a GP object.
        data :
            Creates the dictionary to be passed to `GP.pred` to represent ``y``.
        pred :
            Evaluate the regression function at given locations.

        See also
        --------
        lsqfitgp.BART
        
        References
        ----------
        .. [1] P. Richard Hahn, Jared S. Murray, Carlos M. Carvalho "Bayesian
            Regression Tree Models for Causal Inference: Regularization,
            Confounding, and Heterogeneous Effects (with Discussion)," Bayesian
            Analysis 15(3), 965-1056, September 2020,
            https://doi.org/10.1214/19-BA1195
        """

        # convert covariates to StructuredArray
        x_mu = self._to_structured(x_mu)
        if x_tau is not None:
            x_tau = self._to_structured(x_tau)
            assert x_tau.shape == x_mu.shape
        if x_aux is not None:
            x_aux = self._to_structured(x_aux)
            assert x_aux.shape == x_mu.shape
    
        # convert outcomes, treatment, propensity score, weights to 1d arrays
        y = self._to_vector(y)
        z = self._to_vector(z)
        pihat = self._to_vector(pihat)
        assert y.shape == z.shape == pihat.shape == x_mu.shape
        if weights is not None:
            weights = self._to_vector(weights)
            assert weights.shape == x_mu.shape

        # check include_pi
        if include_pi not in ('mu', 'tau', 'both'):
            raise KeyError(f'invalid value include_pi={include_pi!r}')
        self._include_pi = include_pi

        # add pihat to covariates
        x_mu, x_tau = self._append_pihat(x_mu, x_tau, pihat)
    
        # grid and indices
        splits_mu = _kernels.BART.splits_from_coord(x_mu)
        i_mu = self._toindices(x_mu, splits_mu)
        if x_tau is None:
            splits_tau = splits_mu
            i_tau = None
        else:
            splits_tau = _kernels.BART.splits_from_coord(x_tau)
            i_tau = self._toindices(x_tau, splits_tau)

        # data-dependent prior parameters
        ymin = jnp.min(y)
        ymax = jnp.max(y)
        ystd = jnp.std(y)
        mu_mu = (ymax + ymin) / 2
        k_sigma_mu = (ymax - ymin) / 2
        squares = (y - y.mean()) ** 2
        if weights is not None:
            squares *= weights
        sigma2_priormean = numpy.mean(squares)

        # prior on hyperparams
        hyperprior = copula.makedict({
            'm': gvar.gvar(mu_mu, k_sigma_mu),
            'log(sigma^2)': gvar.gvar(numpy.log(sigma2_priormean), 2),
            'lambda_mu': copula.halfcauchy(2),
            'lambda_tau': copula.halfnorm(1.48),
            'alpha_mu': copula.beta(2, 1),
            'alpha_tau': copula.beta(2, 1),
            'beta_mu': copula.invgamma(1, 1),
            'beta_tau': copula.invgamma(1, 1),
            'z_0': copula.uniform(0, 1),
        })
        if marginalize_mean:
            hyperprior.pop('m')

        # add user hyperparams
        otherhp = gvar.BufferDict(otherhp)
        for key in otherhp.all_keys():
            if hyperprior.has_dictkey(key):
                raise ValueError(f'otherhp[{key!r}] overrides another hyperparameter')
        hyperprior.update(otherhp)

        # GP factory
        def makegp(hp, *, z, i_mu, i_tau, pihat, x_aux, weights,
            splits_mu, splits_tau, ystd, k_sigma_mu, **_):
            
            # TODO maybe I should pass kernelkw_* as arguments, but they may not
            # be jittable. I need jitkw in empbayes_fit for that.

            kw_overridable = dict(
                maxd=10,
                reset=[2, 4, 6, 8],
                gamma=0.95,
                intercept=False,
            )
            kw_not_overridable = dict(indices=True)

            gp = _GP.GP(checkpos=False, checksym=False, solver='chol')

            kernelkw = dict(mu=kernelkw_mu, tau=kernelkw_tau)
                # eval() won't work because they are not captured elsewhere
            
            for name in 'mu', 'tau':
                kw = dict(
                    alpha=hp[f'alpha_{name}'],
                    beta=hp[f'beta_{name}'],
                    dim=name,
                    splits=eval(f'splits_{name}'),
                    **kw_overridable,
                )
                kw.update(kernelkw[name])
                kernel = _kernels.BART(**kw, **kw_not_overridable)
                kernel *= (hp[f'lambda_{name}'] * ystd) ** 2
                
                gp = gp.defproc(name, kernel)

            if 'm' in hp:
                kernel_mean = 0 * _kernels.Constant()
            else:
                kernel_mean = k_sigma_mu ** 2 * _kernels.Constant()
            gp = gp.defproc('m', kernel_mean)

            if gpaux is None:
                gp = gp.defproc('aux', 0 * _kernels.Constant())
            else:
                gp = gpaux(hp, gp)

            gp = gp.deflintransf(
                gp.DefaultProcess,
                lambda m, mu, tau, aux: lambda x:
                m(x) + mu(x) + tau(x) * (x['z'] - hp['z_0']) + aux(x),
                ['m', 'mu', 'tau', 'aux'],
            )
            
            x = self._join_points(True, z, i_mu, i_tau, pihat, x_aux)
            gp = gp.addx(x, 'trainmean')
            errcov = self._error_cov(hp, weights, x)
            return (gp
                .addcov(errcov, 'trainnoise')
                .addtransf({'trainmean': 1, 'trainnoise': 1}, 'train')
            )

        # data factory
        def info(hp, *, y, mu_mu, **_):
            return {'train': y - hp.get('m', mu_mu)}

        # fit hyperparameters
        gpkw = dict(
            y=y,
            z=z,
            i_mu=i_mu,
            i_tau=i_tau,
            pihat=pihat,
            x_aux=x_aux,
            weights=weights,
            splits_mu=splits_mu,
            splits_tau=splits_tau,
            ystd=ystd,
            mu_mu=mu_mu,
            k_sigma_mu=k_sigma_mu,
        )
        options = dict(
            verbosity=3,
            minkw=dict(method='l-bfgs-b', options=dict(maxls=4, maxiter=100)),
            mlkw=dict(epsrel=0),
            forward=True,
            gpfactorykw=gpkw,
        )
        options.update(fitkw)
        fit = _fit.empbayes_fit(hyperprior, makegp, info, **options)
        
        # extract hyperparameters from minimization result
        self.m = fit.p.get('m', mu_mu)
        self.sigma = gvar.sqrt(fit.p['sigma^2'])
        self.sdev_mu = fit.p['lambda_mu'] * ystd
        self.sdev_tau = fit.p['lambda_tau'] * ystd
        self.alpha_mu = fit.p['alpha_mu']
        self.alpha_tau = fit.p['alpha_tau']
        self.beta_mu = fit.p['beta_mu']
        self.beta_tau = fit.p['beta_tau']
        self.z_0 = fit.p['z_0']

        # save fit object
        self.fit = fit

    def _append_pihat(self, x_mu, x_tau, pihat):
        ip = self._include_pi
        if ip == 'mu' or ip == 'both':
            x_mu = _array.StructuredArray.from_dict(dict(
                x=x_mu,
                pihat=pihat,
            ))
        if x_tau is not None and (ip == 'tau' or ip == 'both'):
            x_tau = _array.StructuredArray.from_dict(dict(
                x=x_tau,
                pihat=pihat,
            ))
        return x_mu, x_tau

    def _join_points(self, train, z, i_mu, i_tau, pihat, x_aux):
        """ join covariates into a single StructuredArray """
        columns = dict(
            train=jnp.broadcast_to(bool(train), z.shape),
            i=jnp.arange(z.size).reshape(z.shape),
            z=z,
            mu=i_mu,
            tau=i_mu if i_tau is None else i_tau,
            pihat=pihat,
        )
        if x_aux is not None:
            columns.update(aux=x_aux)
        return _array.StructuredArray.from_dict(columns)

    @staticmethod
    def _error_cov(hp, weights, x):
        """ fill error covariance matrix """
        if weights is None:
            error_var = jnp.broadcast_to(hp['sigma^2'], len(x))
        else:
            error_var = hp['sigma^2'] / weights
        return jnp.diag(error_var)

    def _gethp(self, hp, rng):
        if not isinstance(hp, str):
            return hp
        elif hp == 'map':
            return self.fit.pmean
        elif hp == 'sample':
            return _fastraniter.sample(self.fit.pmean, self.fit.pcov, rng=rng)
        else:
            raise KeyError(hp)

    def gp(self, *, hp='map', z=None, x_mu=None, x_tau=None, pihat=None,
        x_aux=None, weights=None, rng=None):
        """
        Create a Gaussian process with the fitted hyperparameters.

        Parameters
        ----------
        hp : str or dict
            The hyperparameters to use. If ``'map'`` (default), use the marginal
            maximum a posteriori. If ``'sample'``, sample hyperparameters from
            the posterior. If a dict, use the given hyperparameters.
        z : (m,) array, series or dataframe, optional
            Treatment status at test points. If specified, also `x_mu`, `pihat`,
            `x_tau` and `x_aux` (the latter two if and only also specified at
            initialization) must be specified.
        x_mu : (m, p) array, series or dataframe, optional
            Control model covariates at test points.
        x_tau : (m, q) array, series or dataframe, optional
            Moderating model covariates at test points.
        pihat : (m,) array, series or dataframe, optional
            Estimated propensity score at test points.
        x_aux : (m, k) array, series or dataframe, optional
            Additional covariates for the ``'aux'`` process.
        weights : (m,) array, series or dataframe, optional
            Weights for the error variance on the test points.
        rng : numpy.random.Generator, optional
            Random number generator, used if ``hp == 'sample'``.

        Returns
        -------
        gp : GP
            A centered Gaussian process object. To add the mean, use the `m`
            attribute of the `bcf` object. The keys of the GP are ``'Xmean'``,
            ``'Xnoise'``, and ``'X'``, where the "X" stands either for 'train'
            or 'test', and X = Xmean + Xnoise.
        """

        hp = self._gethp(hp, rng)
        return self._gp(hp, z, x_mu, x_tau, pihat, x_aux, weights, self.fit.gpfactorykw)

    def _gp(self, hp, z, x_mu, x_tau, pihat, x_aux, weights, gpfactorykw):
        """
        Internal function to create the GP object. This function must work
        both if the arguments are user-provided and need to be checked and
        converted to standard format, or if they are traced jax values.
        """

        # create GP object
        gp = self.fit.gpfactory(hp, **gpfactorykw)

        # add test points
        if z is not None:

            # check presence/absence of arguments is coherent
            self._check_coherent_covariates(z, x_mu, x_tau, pihat, x_aux)

            # check treatment and propensity score
            z = self._to_vector(z)
            pihat = self._to_vector(pihat)
            assert pihat.shape == z.shape

            # check weights
            if weights is not None:
                weights = self._to_vector(weights)
                assert weights.shape == z.shape

            # add propensity score to covariates
            x_mu = self._to_structured(x_mu)
            assert x_mu.shape == z.shape
            if x_tau is not None:
                x_tau = self._to_structured(x_tau)
                assert x_tau.shape == z.shape
            x_mu, x_tau = self._append_pihat(x_mu, x_tau, pihat)

            # convert covariates to indices
            i_mu = self._toindices(x_mu, gpfactorykw['splits_mu'])
            assert i_mu.dtype == gpfactorykw['i_mu'].dtype
            if x_tau is not None:
                i_tau = self._toindices(x_tau, gpfactorykw['splits_tau'])
                assert i_tau.dtype == gpfactorykw['i_tau'].dtype
            else:
                i_tau = None

            # check auxiliary points
            if x_aux is not None:
                x_aux = self._to_structured(x_aux)

            # add test points
            x = self._join_points(False, z, i_mu, i_tau, pihat, x_aux)
            gp = gp.addx(x, 'testmean')
            errcov = self._error_cov(hp, weights, x)
            gp = (gp
                .addcov(errcov, 'testnoise')
                .addtransf({'testmean': 1, 'testnoise': 1}, 'test')
            )

        return gp

    def _check_coherent_covariates(self, z, x_mu, x_tau, pihat, x_aux):
        if z is None:
            assert x_mu is None
            assert x_tau is None
            assert pihat is None
            assert x_aux is None
        else:
            assert x_mu is not None
            assert pihat is not None
            train_tau = self.fit.gpfactorykw['i_tau']
            if x_tau is None:
                assert train_tau is None
            else:
                assert train_tau is not None
            train_aux = self.fit.gpfactorykw['x_aux']
            if x_aux is None:
                assert train_aux is None
            else:
                assert train_aux is not None

    def data(self, *, hp='map', rng=None):
        """
        Get the data to be passed to `GP.pred` on a GP object returned by `gp`.

        Parameters
        ----------
        hp : str or dict
            The hyperparameters to use. If ``'map'`` (default), use the marginal
            maximum a posteriori. If ``'sample'``, sample hyperparameters from
            the posterior. If a dict, use the given hyperparameters.
        rng : numpy.random.Generator, optional
            Random number generator, used if ``hp == 'sample'``.

        Returns
        -------
        data : dict
            A dictionary representing ``y`` in the format required by the
            `GP.pred` method.
        """

        hp = self._gethp(hp, rng)
        return self.fit.data(hp, **self.fit.gpfactorykw)

    def pred(self, *, hp='map', error=False, format='matrices', z=None,
        x_mu=None, x_tau=None, pihat=None, x_aux=None, weights=None, rng=None):
        r"""
        Predict the outcome at given locations.

        Parameters
        ----------
        hp : str or dict
            The hyperparameters to use. If ``'map'`` (default), use the marginal
            maximum a posteriori. If ``'sample'``, sample hyperparameters from
            the posterior. If a dict, use the given hyperparameters.
        error : bool
            If ``False`` (default), make a prediction for the latent mean. If
            ``True``, add the error term.     
        format : {'matrices', 'gvar'}
            If ``'matrices'`` (default), return the mean and covariance matrix
            separately. If ``'gvar'``, return an array of `~gvar.GVar`.
        z : (m,) array, series or dataframe, optional
            Treatment status at test points. If specified, also `x_mu`, `pihat`,
            `x_tau` and `x_aux` (the latter two if and only also specified at
            initialization) must be specified.
        x_mu : (m, p) array, series or dataframe, optional
            :math:`\mu` model covariates at test points.
        x_tau : (m, q) array, series or dataframe, optional
            :math:`\tau` model covariates at test points.
        pihat : (m,) array, series or dataframe, optional
            Estimated propensity score at test points.
        x_aux : (m, k) array, series or dataframe, optional
            Additional covariates for the ``'aux'`` process at test points.
        weights : (m,) array, series or dataframe, optional
            Weights for the error variance on the test points.
        rng : numpy.random.Generator, optional
            Random number generator, used if ``hp == 'sample'``.

        Returns
        -------
        If `format` is ``'matrices'`` (default):

        mean, cov : arrays
            The mean and covariance matrix of the Normal posterior distribution
            over the regression function at the specified locations.

        If `format` is ``'gvar'``:

        out : array of gvar
            The same distribution represented as an array of `~gvar.GVar`
            objects.
        """
        
        # get hyperparameters
        hp = self._gethp(hp, rng)
        
        # check presence of covariates is coherent
        self._check_coherent_covariates(z, x_mu, x_tau, pihat, x_aux)

        # convert all inputs to arrays compatible with jax to pass them to the
        # compiled implementation
        if z is not None:
            z = self._to_vector(z)
            pihat = self._to_vector(pihat)
            x_mu = self._to_structured(x_mu)
            if x_tau is not None:
                x_tau = self._to_structured(x_tau)
            if x_aux is not None:
                x_aux = self._to_structured(x_aux)
        if weights is not None:
            weights = self._to_vector(weights)
        
        # GP regression
        mean, cov = self._pred(hp, z, x_mu, x_tau, pihat, x_aux, weights, self.fit.gpfactorykw, bool(error))

        # pack output
        if format == 'gvar':
            return gvar.gvar(mean, cov, fast=True)
        elif format == 'matrices':
            return mean, cov
        else:
            raise KeyError(format)

    @functools.cached_property
    def _pred(self):
        
        @functools.partial(jax.jit, static_argnums=(8,))
        def _pred(hp, z, x_mu, x_tau, pihat, x_aux, weights, gpfactorykw, error):
            gp = self._gp(hp, z, x_mu, x_tau, pihat, x_aux, weights, gpfactorykw)
            data = self.fit.data(hp, **gpfactorykw)
            if z is None:
                label = 'train'
            else:
                label = 'test'
            if not error:
                label += 'mean'
            outmean, outcov = gp.predfromdata(data, label, raw=True)
            return outmean + hp.get('m', gpfactorykw['mu_mu']), outcov

        # TODO make everything pure and jit this per class instead of per
        # instance

        return _pred

    @classmethod
    def _to_structured(cls, x, *, check_numerical=True):

        # convert to StructuredArray
        if hasattr(x, 'columns'):
            x = _array.StructuredArray.from_dataframe(x)
        elif hasattr(x, 'to_numpy'):
            x = _array.StructuredArray.from_dict({
                'f0' if x.name is None else x.name: x.to_numpy()
            })
        elif x.dtype.names is None:
            x = _array.unstructured_to_structured(x)
        else:
            x = _array.StructuredArray(x)

        # check fields are numerical, for BART
        if check_numerical:
            assert x.ndim == 1
            assert x.size > len(x.dtype)
            def check_numerical(path, dtype):
                if not numpy.issubdtype(dtype, numpy.number):
                    raise TypeError(f'covariate `{path}` is not numerical')
            cls._walk_dtype(x.dtype, check_numerical)

        return x

    @staticmethod
    def _to_vector(x):
        if hasattr(x, 'columns'): # dataframe
            x = x.to_numpy().squeeze(axis=1)
        elif hasattr(x, 'to_numpy'): # series (dataframe column)
            x = x.to_numpy()
        x = jnp.asarray(x)
        if x.ndim != 1:
            raise ValueError(f'array is not 1d vector, ndim={x.ndim}')
        return x

    @classmethod
    def _walk_dtype(cls, dtype, task, path=None):
        if dtype.names is None:
            task(path, dtype)
        else:
            for name in dtype.names:
                subpath = name if path is None else path + ':' + name
                cls._walk_dtype(dtype[name], task, subpath)

    @staticmethod
    def _toindices(x, splits):
        ix = _kernels.BART.indices_from_coord(x, splits)
        dtype = cast(x.dtype, ix.dtype)
        return _array.unstructured_to_structured(ix, dtype=dtype)

    def __repr__(self):
        if hasattr(self.m, 'sdev'):
            m = str(self.m)
        else:
            m = f'{self.m:.3g}'

        out = f"""BCF fit:
alpha_mu/tau = {self.alpha_mu} {self.alpha_tau} (0 -> intercept only, 1 -> any)
beta_mu/tau = {self.beta_mu} {self.beta_tau} (0 -> any, ∞ -> no interactions)
z_0 = {self.z_0} (mu model applies at z = z_0)
m = {m}
lambda_mu/tau std(y) = {self.sdev_mu} {self.sdev_tau} (large -> conservative extrapolation)
std(y) = {self.fit.gpfactorykw['ystd']:.3g}"""

        weights = self.fit.gpfactorykw['weights']
        if weights is None:
            out += f"""
sigma = {self.sigma}"""
        
        else:
            weights = numpy.array(weights) # to avoid jax taking over the ops
            avgsigma = numpy.sqrt(numpy.mean(self.sigma ** 2 / weights))
            out += f"""
sqrt(mean(sigma^2/w))  = {avgsigma}
sigma = {self.sigma}"""

        return out
